from mxcubecore.BaseHardwareObjects import HardwareObject

from .BeamCmds import (
    ControllerCommand,
    HWObjActuatorCommand,
)


class ID232BeamCmds(HardwareObject):
    def __init__(self, *args):
        HardwareObject.__init__(self, *args)

    def init(self):
        controller = self.get_object_by_role("controller")
        detcover = self.get_object_by_role("detcover")
        scintillator = self.get_object_by_role("scintillator")
        hutchtrigger = self.get_object_by_role("hutchtrigger")
        cryo = self.get_object_by_role("cryo")

        self.centrebeam = ControllerCommand("Centre beam", controller.centrebeam)
        self.anneal = ControllerCommand("Anneal", controller.anneal)

        self.detcover = HWObjActuatorCommand("Detector cover", detcover)
        self.scintillator = HWObjActuatorCommand("Scintillator", scintillator)
        self.hutchtrigger = HWObjActuatorCommand("Hutchtrigger", hutchtrigger)
        self.cryo = HWObjActuatorCommand("Cryo", cryo)

    def get_commands(self):
        return [
            self.centrebeam,
            self.anneal,
            self.detcover,
            self.scintillator,
            self.hutchtrigger,
            self.cryo,
        ]



import logging
import os
import time

import Session

from mxcubecore import HardwareRepository as HWR


class SOLEILSession(Session.Session):
    def __init__(self, *args, **kwargs):
        Session.Session.__init__(self, *args, **kwargs)
        self.username = ""
        self.gid = ""
        self.uid = ""
        self.projuser = ""

    def path_to_ispyb(self, path):
        ispyb_base = self["file_info"].get_property("ispyb_base_directory") % {
            "projuser": self.projuser
        }
        path = path.replace(
            self["file_info"].get_property("base_directory"), ispyb_base
        )
        return path

    def set_user_info(self, username, user_id, group_id, projuser=None):
        logging.debug(
            "SESSION - User %s logged in. gid=%s / uid=%s "
            % (username, group_id, user_id)
        )
        self.username = username
        self.group_id = group_id
        self.user_id = user_id
        self.projuser = projuser

    def get_proposal_number(self):
        """
        :returns: The proposal, 'local-user' if no proposal is
                  available
        :rtype: str
        """

        if self.proposal_number:
            return "%s" % (self.proposal_number)
        else:
            return "local-user"

    def get_base_data_directory(self):
        """
        Returns the base data directory taking the 'contextual'
        information into account, such as if the current user
        is inhouse.

        :returns: The base data path.
        :rtype: str
        """
        user_category = ""
        directory = ""

        if self.session_start_date:
            start_time = self.session_start_date.split(" ")[0]  # .replace('-', '')
        else:
            # PL. To avoid mixing users directory if they restart the application
            # after midnight but before 8 AM, the directory date doesn't change:
            _local_time = time.localtime()
            if _local_time[3] > 7:
                start_time = time.strftime("%Y-%m-%d")
            else:
                # substract 8 hours to current date to get yesterday's date.
                _local_time = time.gmtime((time.time() - 8 * 60 * 60))
                start_time = time.strftime("%Y-%m-%d", _local_time)

        if self.is_inhouse():
            # directory = os.path.join(self.base_directory, self.endstation_name,
            #                         self.get_user_category(), self.get_proposal(),
            #                         start_time)
            directory = os.path.join(
                self.base_directory,
                start_time,
                self.proposal_number,
                self.get_proposal_number(),
            )
        else:
            # directory = os.path.join(self.base_directory, self.get_user_category(),
            #                         self.get_proposal(), self.endstation_name,
            #                         start_time)
            logging.debug("SoleilSession self.base_directory %s" % self.base_directory)
            logging.debug("SoleilSession start_time %s" % start_time)
            logging.debug(
                "SoleilSession self.proposal_number %s" % self.proposal_number
            )
            logging.debug(
                "SoleilSession self.get_proposal_number() %s"
                % self.get_proposal_number()
            )
            directory = os.path.join(
                self.base_directory, start_time, self.get_proposal_number()
            )

        return directory

    # def get_rawdata_directory(self, directory=None):
    #    if directory is None:
    #        thedir = self.get_base_data_directory()
    #    else:
    #        thedir = directory
    #    if 'RAW_DATA' not in thedir:
    #        thedir = os.path.join(thedir, 'ARCHIVE')
    #    return thedir

    def get_archive_directory(self, directory=None):
        if directory is None:
            thedir = self.get_base_data_directory()
        else:
            thedir = directory

        if "RAW_DATA" in thedir:
            thedir = thedir.replace("RAW_DATA", "ARCHIVE")
        else:
            thedir = os.path.join(thedir, "ARCHIVE")

        return thedir

    def get_ruche_info(self, path):

        if self.is_inhouse(self.username):
            usertype = "soleil"
        else:
            usertype = "users"

        basedir = os.path.dirname(path)
        ruchepath = basedir.replace(
            self["file_info"].get_property("base_directory"), ""
        )
        if ruchepath and ruchepath[0] == os.path.sep:
            ruchepath = ruchepath[1:]

        infostr = "%s %s %s %s %s %s\n" % (
            usertype,
            self.username,
            self.user_id,
            self.group_id,
            basedir,
            ruchepath,
        )
        return infostr


def test():
    hwr = HWR.get_hardware_repository()
    hwr.connect()

    sess = HWR.beamline.session

    sess.set_user_info("mx2014", "143301", "14330", "20100023")

    path = "/927bis/ccd/2015_Run2/visitor/mx2014/px2/20150120/ARCHIVE/mx2014/mx2014_2_4.snapshot.jpeg"
    ispyb_path = sess.path_to_ispyb(path)

    print(path)
    print("  will become ")
    print(ispyb_path)

    # print sess.get_ruche_info(path)


if __name__ == "__main__":
    test()



"""
Plate Manipulator maintenance.
"""
from mxcubecore.BaseHardwareObjects import HardwareObject


class PlateManipulatorMaintenance(HardwareObject):

    __TYPE__ = "PlateManipulatorMaintenance"

    """
    Actual implementation of the Plate Manipulator MAINTENANCE,
    COMMANDS ONLY
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def init(self):
        self._sc = self.get_object_by_role("sample_changer")
        self._scan_limits = ''

    def _do_abort(self):
        """
        Abort current command

        :returns: None
        :rtype: None
        """
        return self._sc._do_abort()
    
    def _move_to_crystal_position(self, args):
        """
        command to move MD head to x, y crystal position
        argrs: crystall uuid or none for current drop position

        :returns: None if exception
        :rtype: None
        """
        return self._sc.move_to_crystal_position(args)
    

    def _do_change_mode(self, args):
        self._sc._do_change_mode(args)

    def set_plate_barcode(self, args):
        """
        Plate barcode is set in config, but can be update here
        :returns: None
        :rtype: None
        """
        ret = self._sc.change_plate_barcode(args)
        if ret:
            self._update_global_state()

    def _get_scan_limits(self, args):
        """
        Omega Dynamic scan limit current command
        :returns: None
        :rtype: None
        """
        self._scan_limits = self._sc.get_scan_limits(args)
        self._update_global_state()


    def _update_global_state(self):
        state_dict, cmd_state, message = self.get_global_state()
        self.emit("globalStateChanged", (state_dict, cmd_state, message))
    
    def get_global_state(self):
        """
        """
        state = self._sc._read_state()
        scan_limits = self._scan_limits
        # ready = self._sc._ready()
        running = state in ("RUNNING",)
        plate_info_dict =  self._sc.get_plate_info()
        state_dict = {
            "running": running,
            "scan_limits": scan_limits,
            "state": state,
            "plate_info" : plate_info_dict
        }

        cmd_state = {
            "abort": True,
        }

        message = ""

        return state_dict, cmd_state, message

    

    def get_cmd_info(self):
        """ return information about existing commands for this object
           the information is organized as a list
           with each element contains
           [ cmd_name,  display_name, category ]
        """
        """ [cmd_id, cmd_display_name, nb_args, cmd_category, description ] """

        cmd_list = [
            [
                "Actions",
                [
                    ["abort", "Abort", "Actions", None],
                ],
            ],
        ]

        return cmd_list

    def send_command(self, cmdname, args=None):
        if cmdname in ["getOmegaMotorDynamicScanLimits"]:
            self._get_scan_limits(args)
        if cmdname in ["moveToCrystalPosition"]:
            self._move_to_crystal_position(args)
        if cmdname == "abort":
            self._do_abort()
        if cmdname == "setPlateBarcode":
            self.set_plate_barcode(args)
        # if cmdname == "change_mode":
        #     self._do_change_mode(args)      
        return True



#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""Abstract Online Processing class"""
import json
import logging
import os
import subprocess
import time
from copy import copy

import gevent
import matplotlib.pyplot as plt
import numpy as np
import SimpleHTML
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy import ndimage
from scipy.interpolate import UnivariateSpline

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject

__copyright__ = """ Copyright © 2010-2022 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


DEFAULT_RESULT_TYPES = [
    {"key": "spots_resolution", "descr": "Resolution", "color": (120, 0, 0)},
    {"key": "score", "descr": "Score", "color": (0, 120, 0)},
    {"key": "spots_num", "descr": "Number of spots", "color": (0, 0, 120)},
    {"key": "is", "descr": "Intensity", "color": (0, 0, 120)},
]


"""
AbstractOnlineProcessing hardware object handles online data processing.
Typical example of online processing is a mesh scan where user is provided
with real-time results describing diffraction quality.
Method run_processing is called from the queue_entry when the data collection
starts. Then empty arrays to store results are created.
Typicaly an input file is created and processing is started with script via
subprocess.Popen. Results are emited with onlineProcessingResults signal.

Implementations:
 * DozorOnlinelProcessing: online processing based on the EDNA Dozor plugin.
   Started with EDNA and results are set via xmlrpc.
 * OnlineProcessigMockup: mockup version allows to simulate online processing.
"""


class AbstractOnlineProcessing(HardwareObject):
    def __init__(self, name):
        super().__init__(name)

        # Hardware objects ----------------------------------------------------
        self.ssx_setup = None

        # Internal variables --------------------------------------------------
        self.start_command = None
        self.kill_command = None
        self.data_collection = None
        self.grid = None
        self.params_dict = None
        self.result_types = None
        self.results_raw = None
        self.results_aligned = None
        self.interpolate_results = None
        self.done_event = None
        self.started = None
        self.workflow_info = None

        self.current_grid_index = None
        self.grid_properties = []

    def init(self):
        self.done_event = gevent.event.Event()
        self.ssx_setup = self.get_object_by_role("ssx_setup")

        self.result_types = self.get_property("result_types", DEFAULT_RESULT_TYPES)
        self.start_command = str(self.get_property("processing_command"))
        self.kill_command = str(self.get_property("kill_command"))
        self.interpolate_results = self.get_property("interpolate_results")

    def get_result_types(self):
        return self.result_types

    def prepare_processing(self):
        """Prepares processing parameters, creates empty result arrays and
           create necessary directories to store results.

        :param data_collection: data collection object
        :type : queue_model_objects.DataCollection
        """
        acquisition = self.data_collection.acquisitions[0]
        acq_params = acquisition.acquisition_parameters
        self.grid = self.data_collection.grid

        grid_params = None
        if self.grid:
            grid_params = self.grid.get_properties()

        prefix = acquisition.path_template.get_prefix()
        run_number = acquisition.path_template.run_number
        first_image_num = acq_params.first_image
        images_num = acq_params.num_images
        last_image_num = first_image_num + images_num - 1
        lines_num = acq_params.num_lines

        template = os.path.join(
            acquisition.path_template.directory,
            "%s_%%d_%%0%sd.cbf"
            % (
                acquisition.path_template.get_prefix(),
                acquisition.path_template.precision,
            ),
        )
        if acquisition.path_template.compression:
            template += ".gz"

        workflow_step_directory = None
        self.params_dict = {}

        if self.data_collection.run_online_processing == "XrayCentering":
            prefix = "xray_centering_%s" % prefix
            if self.grid:
                workflow_step_directory = "/mesh"
            else:
                workflow_step_directory = "/line"

        if self.workflow_info is not None:
            process_directory = self.workflow_info["process_root_directory"]
            archive_directory = self.workflow_info["archive_root_directory"]
        else:
            i = 1
            while True:
                process_input_file_dirname = f"{prefix}_run{run_number}_{i}"
                process_directory = os.path.join(
                    acquisition.path_template.process_directory,
                    process_input_file_dirname,
                )
                archive_directory = os.path.join(
                    acquisition.path_template.get_archive_directory(),
                    process_input_file_dirname,
                )
                if not os.path.exists(process_directory):
                    break
                i += 1

        self.params_dict["process_root_directory"] = process_directory
        self.params_dict["archive_root_directory"] = archive_directory
        self.params_dict["result_file_path"] = archive_directory
        self.params_dict["collection_id"] = HWR.beamline.collect.collection_id

        if workflow_step_directory:
            process_directory += workflow_step_directory
            archive_directory += workflow_step_directory

        try:
            if not os.path.isdir(process_directory):
                os.makedirs(process_directory)
        except Exception:
            logging.getLogger("GUI").exception(
                "Online processing: Unable to create processing directory %s"
                % process_directory
            )
            self.set_processing_status("Failed")

        self.params_dict["process_directory"] = process_directory
        self.params_dict["archive_directory"] = archive_directory

        # self.params_dict["plot_path"] = os.path.join(
        # self.params_dict["directory"],
        #    "online_processing_result.png")

        self.params_dict["folder_path"] = archive_directory
        self.params_dict["snapshot_path"] = os.path.join(
            archive_directory, "snapshot.png"
        )
        self.params_dict["cartography_path"] = os.path.join(
            archive_directory, "online_processing_plot.png"
        )
        self.params_dict["log_file_path"] = os.path.join(
            archive_directory, "online_processing.log"
        )
        self.params_dict["html_file_path"] = os.path.join(
            archive_directory, "index.html"
        )
        self.params_dict["json_file_path"] = os.path.join(
            archive_directory, "report.json"
        )
        self.params_dict["csv_file_path"] = os.path.join(
            archive_directory, "online_processing_results.csv"
        )

        self.params_dict["template"] = template
        self.params_dict["first_image_num"] = first_image_num
        self.params_dict["images_num"] = images_num
        self.params_dict["lines_num"] = lines_num
        self.params_dict["images_per_line"] = images_num / lines_num
        self.params_dict["run_number"] = run_number
        self.params_dict["osc_midle"] = float(acq_params.osc_start)
        self.params_dict["osc_range"] = acq_params.osc_range
        self.params_dict["resolution"] = acq_params.resolution
        self.params_dict["exp_time"] = acq_params.exp_time
        self.params_dict["hare_num"] = acq_params.hare_num

        if not acq_params.num_images_per_trigger:
            self.params_dict["num_images_per_trigger"] = 1
        else:
            self.params_dict["num_images_per_trigger"] = (
                acq_params.num_images_per_trigger
            )

        self.params_dict["status"] = "Started"
        self.params_dict["title"] = "%s_%d_#####.cbf (%d - %d)" % (
            prefix,
            run_number,
            first_image_num,
            last_image_num,
        )
        self.params_dict["comments"] = "Scan lines: %d, frames per line: %d" % (
            lines_num,
            images_num / lines_num,
        )
        self.params_dict["workflow_type"] = self.data_collection.run_online_processing
        self.params_dict["group_id"] = self.data_collection.lims_group_id
        self.params_dict["processing_start_time"] = time.strftime("%Y-%m-%d %H:%M:%S")

        if grid_params:
            self.params_dict["dx_mm"] = grid_params["dx_mm"]
            self.params_dict["dy_mm"] = grid_params["dy_mm"]
            self.params_dict["steps_x"] = grid_params["steps_x"]
            self.params_dict["steps_y"] = grid_params["steps_y"]
            self.params_dict["xOffset"] = grid_params["xOffset"]
            self.params_dict["yOffset"] = grid_params["yOffset"]
            self.params_dict["reversing_rotation"] = grid_params["reversing_rotation"]
            # self.store_coordinate_map()
        else:
            self.params_dict["steps_y"] = 1
            self.params_dict["reversing_rotation"] = False

        self.results_raw = {}
        self.results_aligned = {}

        # Empty numpy arrays to store raw and aligned results
        for result_type in self.result_types:
            # images_num =  self.params_dict["images_num"]
            if "size" in result_type:
                images_num = result_type["size"]
            else:
                images_num = self.params_dict["images_num"]

            self.results_raw[result_type["key"]] = np.zeros(images_num)
            self.results_aligned[result_type["key"]] = np.zeros(images_num)

            if self.interpolate_results:
                self.results_aligned["interp_" + result_type] = np.zeros(images_num)
            if (
                self.data_collection.is_mesh()
                and images_num == self.params_dict["images_num"]
            ):
                self.results_aligned[result_type["key"]] = self.results_aligned[
                    result_type["key"]
                ].reshape(self.params_dict["steps_x"], self.params_dict["steps_y"])

        # if not self.data_collection.is_mesh():
        #    self.results_raw["x_array"] = np.linspace(
        #        0, images_num, images_num, dtype=np.int32
        #    )

        try:
            gevent.spawn(self.save_snapshot_task, self.params_dict["snapshot_path"])
        except Exception:
            logging.getLogger("GUI").exception(
                "Online processing: Could not save snapshot: %s"
                % self.params_dict["snapshot_path"]
            )

        self.emit(
            "processingStarted",
            (self.data_collection, self.results_raw, self.results_aligned),
        )

    def create_processing_input_file(self, processing_input_filename):
        """Creates processing input file
        Args:
            processing_input_filename (str): file name.
        """
        return

    def run_processing(self, data_collection):
        """Starts Online processing
        Args:
            data_collection(queue_model_objects.DataCollection): data collection object
        """
        self.data_collection = data_collection
        self.prepare_processing()
        input_filename = os.path.join(
            self.params_dict["process_directory"], "dozor_input.xml"
        )
        self.create_processing_input_file(input_filename)

        # results = {"raw" : self.results_raw,
        #           "aligned": self.results_aligned}
        # print "emit ! ", results
        # self.emit("processingStarted", (data_collection, results))
        # self.emit("processingResultsUpdate", False)

        if not os.path.isfile(self.start_command):
            msg = (
                f"OnlineProcessing: Start command {self.start_command} "
                + "is not executable"
            )
            logging.getLogger("queue_exec").error(msg)
            self.set_processing_status("Failed")
        else:
            line_to_execute = (
                self.start_command
                + " "
                + input_filename
                + " "
                + self.params_dict["process_directory"]
            )

            self.started = True
            subprocess.Popen(
                str(line_to_execute),
                shell=True,
                stdin=None,
                stdout=None,
                stderr=None,
                close_fds=True,
            )

    def save_snapshot_task(self, snapshot_filename):
        """Saves snapshot
        Args:
            snapshot_filename (str): filename
        """
        try:
            HWR.beamline.collect._take_crystal_snapshot(snapshot_filename)
            logging.getLogger("HWR").info(
                f"Online processing: Snapshot {snapshot_filename} saved."
            )
        except Exception:
            logging.getLogger("GUI").exception(
                f"Online processing: Could not save snapshot {snapshot_filename}"
            )

    def is_running(self):
        """Returns True if processing is running"""
        return not self.done_event.is_set()

    def stop_processing(self):
        """Stops processing"""
        self.started = False
        self.set_processing_status("Stopped")
        # subprocess.Popen(self.kill_command, shell=True, stdin=None,
        #                 stdout=None, stderr=None, close_fds=True)

    def set_processing_status(self, status):
        """Sets processing status and finalize the processing
           Method called from EDNA via xmlrpc
        Args:
            status(str): processing status (Success, Failed)
        """
        self.emit("processingResultsUpdate", True)

        self.data_collection.set_online_processing_results(
            copy(self.results_raw), copy(self.results_aligned)
        )

        if self.params_dict["workflow_type"] == "XrayCentering":
            if self.results_aligned["best_positions"]:
                logging.getLogger("GUI").info(
                    "Xray centering: Moving to the best position"
                )
                HWR.beamline.diffractometer.move_motors(
                    self.results_aligned["center_mass"], timeout=15
                )
                logging.getLogger("GUI").info(
                    "Xray centering: Storing mesh results in ISPyB"
                )
                self.store_processing_results(status)
            else:
                logging.getLogger("GUI").warning(
                    "Xray Centering: No diffraction found. Stopping Xray centering"
                )
                status = "Failed"
                self.workflow_info = None
            self.done_event.set()
            if status == "Failed":
                self.emit("processingFailed")
            else:
                self.emit("processingFinished")
        else:
            self.workflow_info = None
            if status == "Failed":
                self.emit("processingFailed")
            else:
                self.emit("processingFinished")

            self.done_event.set()
            self.store_processing_results(status)

    def store_processing_results(self, status):
        """Stores result plots. In the case of MeshScan and XrayCentering
           html is created and results saved in ISPyB

        :param status: status type
        :type status: str
        """
        log = logging.getLogger("HWR")

        self.started = False
        self.params_dict["status"] = status

        # ---------------------------------------------------------------------
        # Assembling all file names
        self.params_dict["max_dozor_score"] = float(self.results_aligned["score"].max())
        best_positions = self.results_aligned.get("best_positions", [])

        processing_grid_overlay_file = os.path.join(
            self.params_dict["archive_directory"], "grid_overlay.png"
        )

        # If MeshScan and XrayCentring then info is stored in ISPyB
        if self.params_dict["workflow_type"] in (
            "MeshScan",
            "XrayCentering",
            "LineScan",
        ):
            if self.workflow_info is not None:
                self.params_dict["workflow_id"] = self.workflow_info["workflow_id"]

            (
                workflow_id,
                workflow_mesh_id,
                grid_info_id,
            ) = HWR.beamline.lims.store_workflow(self.params_dict)

            self.params_dict["workflow_id"] = workflow_id
            self.params_dict["workflow_mesh_id"] = workflow_mesh_id
            self.params_dict["grid_info_id"] = grid_info_id

            if self.params_dict["workflow_type"] == "XrayCentering" and self.grid:
                self.workflow_info = {
                    "workflow_id": self.params_dict["workflow_id"],
                    "process_root_directory": self.params_dict[
                        "process_root_directory"
                    ],
                    "archive_root_directory": self.params_dict[
                        "archive_root_directory"
                    ],
                }
            else:
                self.workflow_info = None

            HWR.beamline.collect.update_lims_with_workflow(
                workflow_id, self.params_dict["snapshot_path"]
            )

            HWR.beamline.lims.store_workflow_step(self.params_dict)
            if len(best_positions) > 0:
                HWR.beamline.collect._store_image_in_lims_by_frame_num(
                    best_positions[0]["index"]
                )
            log.info("Online processing: Results saved in ISPyB")

        HWR.beamline.lims.set_image_quality_indicators_plot(
            HWR.beamline.collect.collection_id,
            self.params_dict["cartography_path"],
            self.params_dict["csv_file_path"],
        )

        if self.params_dict["lines_num"] > 1:
            fig, ax = plt.subplots(nrows=1, ncols=1)
            current_max = max(fig.get_size_inches())
            grid_width = self.params_dict["steps_x"] * self.params_dict["xOffset"]
            grid_height = self.params_dict["steps_y"] * self.params_dict["yOffset"]

            if grid_width > grid_height:
                fig.set_size_inches(current_max, current_max * grid_height / grid_width)
            else:
                fig.set_size_inches(current_max * grid_width / grid_height, current_max)

            im = ax.imshow(
                np.transpose(self.results_aligned["score"]),
                interpolation="none",
                aspect="auto",
                extent=[
                    0,
                    self.results_aligned["score"].shape[0],
                    0,
                    self.results_aligned["score"].shape[1],
                ],
            )
            im.set_cmap("hot")

            try:
                if not os.path.exists(os.path.dirname(processing_grid_overlay_file)):
                    os.makedirs(os.path.dirname(processing_grid_overlay_file))

                plt.imsave(
                    processing_grid_overlay_file,
                    np.transpose(self.results_aligned["score"]),
                    format="png",
                    cmap="hot",
                )
                self.grid.set_overlay_pixmap(processing_grid_overlay_file)
                log.info(
                    "Online processing: Grid overlay figure saved %s"
                    % processing_grid_overlay_file
                )
            except Exception:
                log.exception(
                    "Online processing: Could not save grid overlay figure %s"
                    % processing_grid_overlay_file
                )

            if len(best_positions) > 0:
                plt.axvline(x=best_positions[0]["col"], linewidth=0.5)
                plt.axhline(y=best_positions[0]["row"], linewidth=0.5)

                divider = make_axes_locatable(ax)
                cax = divider.append_axes("right", size=0.1, pad=0.05)
                cax.tick_params(axis="x", labelsize=8)
                cax.tick_params(axis="y", labelsize=8)
                plt.colorbar(im, cax=cax)
        else:
            fig, ax = plt.subplots(nrows=2, ncols=1)
            max_resolution = self.params_dict["resolution"]
            min_resolution = self.results_aligned["spots_resolution"].max()
            max_score = self.results_aligned["score"].max()

            if max_score == 0:
                max_score = 1
            max_spots_num = self.results_aligned["spots_num"].max()
            if max_spots_num == 0:
                max_spots_num = 1

            ax[0].plot(
                self.results_aligned["score"] / max_score, ",", label="Score", c="r"
            )
            ax[0].plot(
                self.results_aligned["spots_num"] / max_spots_num,
                ",",
                label="Number of spots",
                c="b",
            )
            ax[0].plot(
                self.results_aligned["spots_resolution"], ".", label="Resolution", c="y"
            )

            ax[0].legend(
                loc="lower center",
                fancybox=True,
                numpoints=1,
                borderaxespad=0.0,
                # bbox_to_anchor=(0.5, -0.13),
                ncol=3,
                fontsize=8,
            )
            ax[0].set_ylim(-0.01, 1.1)
            ax[0].set_xlim(0, self.params_dict["images_num"])

            positions = np.linspace(
                0, self.results_aligned["spots_resolution"].max(), 5
            )
            labels = ["inf"]
            for item in positions[1:]:
                labels.append("%.2f" % (1.0 / item))
            ax[0].set_yticks(positions)
            ax[0].set_yticklabels(labels)

            # new_labels = numpy.linspace(min_resolution, max_resolution / 1.2, len(ax.get_yticklabels()))
            # new_labels = numpy.round(new_labels, 1)
            # ax.set_yticklabels(new_labels)
            ax[0].set_ylabel("Resolution")

            ay1 = ax[0].twinx()
            new_labels = np.linspace(
                0,
                self.results_aligned["spots_num"].max(),
                len(ay1.get_yticklabels()),
                dtype=np.int16,
            )
            ay1.set_yticklabels(new_labels)
            ay1.set_ylabel("Number of spots")

            ax[1].plot(self.results_raw["is"], ",", label="Intensity", c="g")
            ax[1].set_ylabel("Intensity")

            for ax_plot in ax:
                ax_plot.tick_params(axis="x", labelsize=8)
                ax_plot.tick_params(axis="y", labelsize=8)
                ax_plot.grid(True)

        # ---------------------------------------------------------------------
        # Stores plot in the processing directory
        try:
            if not os.path.exists(
                os.path.dirname(self.params_dict["cartography_path"])
            ):
                os.makedirs(os.path.dirname(self.params_dict["cartography_path"]))
            fig.savefig(
                self.params_dict["cartography_path"], dpi=100, bbox_inches="tight"
            )
            log.info(
                "Online processing: Plot saved in %s"
                % self.params_dict["cartography_path"]
            )
        except Exception:
            log.exception(
                "Online processing: Could not save plot in %s"
                % self.params_dict["cartography_path"]
            )

        # ---------------------------------------------------------------------
        # Stores plot for ISPyB
        try:
            if not os.path.exists(
                os.path.dirname(self.params_dict["cartography_path"])
            ):
                os.makedirs(os.path.dirname(self.params_dict["cartography_path"]))
            fig.savefig(
                self.params_dict["cartography_path"], dpi=100, bbox_inches="tight"
            )
            log.info(
                "Online processing: Plot for ISPyB saved in %s"
                % self.params_dict["cartography_path"]
            )
        except Exception:
            log.exception(
                "Online processing: Could not save plot for ISPyB %s"
                % self.params_dict["cartography_path"]
            )

        plt.close(fig)

        # ---------------------------------------------------------------------
        # Generates html and json files
        try:
            SimpleHTML.generate_online_processing_report(
                self.results_aligned, self.params_dict
            )
            log.info(
                "Online processing: Html report saved in %s"
                % self.params_dict["html_file_path"]
            )
            log.info(
                "Online processing: Json report saved in %s"
                % self.params_dict["json_file_path"]
            )
        except Exception as ex:
            log.exception(
                "Online processing: Could not save results html %s: %s"
                % (self.params_dict["html_file_path"], str(ex))
            )
            log.exception(
                "Online processing: Could not save json results in %s : %s"
                % (self.params_dict["json_file_path"], str(ex))
            )

        # ---------------------------------------------------------------------
        # Writes results in the csv file
        try:
            det_pixel_size = HWR.beamline.detector.get_pixel_size()
            with open(self.params_dict["csv_file_path"], "w") as processing_csv_file:
                processing_csv_file.write(
                    "%s,%d,%d,%d,%d,%d,%s,%d,%d\n"
                    % (
                        self.params_dict["template"],
                        self.params_dict["first_image_num"],
                        self.params_dict["images_num"],
                        self.params_dict["run_number"],
                        self.params_dict["run_number"],
                        self.params_dict["lines_num"],
                        str(self.params_dict["reversing_rotation"]),
                        det_pixel_size[0],
                        det_pixel_size[1],
                    )
                )
                for index in range(self.params_dict["images_num"]):
                    processing_csv_file.write(
                        "%d,%f,%d,%f\n"
                        % (
                            index,
                            self.results_raw["score"][index],
                            self.results_raw["spots_num"][index],
                            self.results_raw["spots_resolution"][index],
                        )
                    )
            log.info(
                "Online processing: Raw data stored in %s"
                % self.params_dict["csv_file_path"]
            )
        except Exception:
            log.error(
                "Online processing: Unable to store raw data in %s"
                % self.params_dict["csv_file_path"]
            )
        # ---------------------------------------------------------------------

    def align_processing_results(self, start_index, end_index):
        """Realigns all results. Each results (one dimensional numpy array)
        is converted to 2d numpy array according to diffractometer geometry.
        Function also extracts 10 (if they exist) best positions
        """
        # Each result array is realigned

        for score_key in self.results_raw:
            if (
                self.grid
                and self.results_raw[score_key].size == self.params_dict["images_num"]
            ):
                for cell_index in range(start_index, end_index + 1):
                    col, row = self.grid.get_col_row_from_image_serial(
                        cell_index + self.params_dict["first_image_num"]
                    )
                    if (
                        col < self.results_aligned[score_key].shape[0]
                        and row < self.results_aligned[score_key].shape[1]
                    ):
                        self.results_aligned[score_key][col][row] = self.results_raw[
                            score_key
                        ][cell_index]
            else:
                self.results_aligned[score_key] = self.results_raw[score_key]
                if self.interpolate_results:
                    x_array = np.linspace(
                        0,
                        self.params_dict["images_num"],
                        self.params_dict["images_num"],
                        dtype=int,
                    )
                    spline = UnivariateSpline(
                        x_array, self.results_aligned[score_key], s=10
                    )
                    self.results_aligned["interp_" + score_key] = spline(x_array)

        if self.grid:
            self.grid.set_score(self.results_raw["spots_num"])
            (center_x, center_y) = ndimage.measurements.center_of_mass(
                self.results_aligned["score"]
            )
            self.results_aligned["center_mass"] = self.grid.get_motor_pos_from_col_row(
                center_x, center_y
            )
        else:
            centred_positions = self.data_collection.get_centred_positions()
            if len(centred_positions) == 2:
                center_x = ndimage.measurements.center_of_mass(
                    self.results_aligned["score"]
                )[0]
                self.results_aligned["center_mass"] = (
                    HWR.beamline.diffractometer.get_point_from_line(
                        centred_positions[0],
                        centred_positions[1],
                        center_x,
                        self.params_dict["images_num"],
                    )
                )
            else:
                self.results_aligned["center_mass"] = centred_positions[0]

        # Best positions are extracted
        best_positions_list = []

        index_arr = (-self.results_raw["score"]).argsort()[:10]
        if len(index_arr) > 0:
            for index in index_arr:
                if self.results_raw["score"][index] > 0:
                    best_position = {}
                    best_position["index"] = index
                    best_position["index_serial"] = (
                        self.params_dict["first_image_num"] + index
                    )
                    best_position["score"] = self.results_raw["score"][index]
                    best_position["spots_num"] = self.results_raw["spots_num"][index]
                    best_position["spots_resolution"] = self.results_raw[
                        "spots_resolution"
                    ][index]
                    best_position["filename"] = os.path.basename(
                        self.params_dict["template"]
                        % (
                            self.params_dict["run_number"],
                            self.params_dict["first_image_num"] + index,
                        )
                    )

                    cpos = None
                    if self.grid:
                        col, row = self.grid.get_col_row_from_image_serial(
                            index + self.params_dict["first_image_num"]
                        )
                        col += 0.5
                        row = self.params_dict["steps_y"] - row - 0.5
                        cpos = self.grid.get_motor_pos_from_col_row(col, row)
                    else:
                        col = index
                        row = 0
                        cpos = None
                        # TODO make this nicer
                        # num_images = self.data_collection.acquisitions[0].acquisition_parameters.num_images - 1
                        # (point_one, point_two) = self.data_collection.get_centred_positions()
                        # cpos = HWR.beamline.diffractometer.get_point_from_line(point_one, point_two, index, num_images)
                    best_position["col"] = col
                    best_position["row"] = row
                    best_position["cpos"] = cpos
                    best_positions_list.append(best_position)

        self.results_aligned["best_positions"] = best_positions_list

    def extract_sweeps(self):
        """Extracts sweeps from processing results"""

        # self.results_aligned
        logging.getLogger("HWR").info("Online processing: Extracting sweeps")
        for col in range(self.results_aligned["score"].shape[1]):
            mask = self.results_aligned["score"][:, col] > 0
            label_im, nb_labels = ndimage.label(mask)
            # sizes = ndimage.sum(mask, label_im, range(nb_labels + 1))
            labels = np.unique(label_im)
            label_im = np.searchsorted(labels, label_im)

    def store_coordinate_map(self):
        """Store map"""
        mesh_best_file = os.path.join(
            self.params_dict["process_directory"], "mesh_best.json"
        )

        json_dict = {"meshPositions": []}
        for index, item in enumerate(self.grid.get_coordinate_map()):
            json_dict["meshPositions"].append(
                {"index": index, "indexY": item[0], "indexZ": item[1]}
            )
        with open(mesh_best_file, "w") as fp:
            json.dump(json_dict, fp)

        self.print_log(f"Online processing: Mesh best file {mesh_best_file} saved")



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
BeamDefiner ESRF implementation class - methods to define the size and shape of
the beam.
"""

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


import logging

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractBeam import AbstractBeam


class ESRFBeam(AbstractBeam):
    """Beam ESRF implementation"""

    unit = "mm"

    def init(self):
        """Initialize hardware"""
        super().init()

        _definer_type = []
        self._aperture = self.get_object_by_role("aperture")
        if self._aperture:
            _definer_type.append("aperture")

        _slits = self.get_property("slits")
        if _slits:
            self._slits = {}
            _definer_type.append("slits")
            _bliss_obj = self.get_object_by_role("bliss")
            for name in _slits.split():
                _key, _val = name.split(":")
                self._slits.update({_key: _bliss_obj.getattribute(_val)})

        self._definer = self.get_object_by_role("definer")
        if self._definer:
            _definer_type.append("definer")

        if len(_definer_type) == 1:
            self._definer_type = _definer_type[0]
        else:
            self._definer_type = None

        self._definer_type = self.get_property("definer_type") or self._definer_type

        beam_position = self.get_property("beam_position")

        if beam_position:
            self._beam_position_on_screen = tuple(map(float, beam_position.split()))

        if self._aperture:
            self._aperture.connect("valueChanged", self._re_emit_values)
            self._aperture.connect("stateChanged", self._re_emit_values)

        if self._definer:
            self._definer.connect("valueChanged", self._re_emit_values)
            self._definer.connect("stateChanged", self._re_emit_values)

    def _re_emit_values(self, value):
        # redefine as re_emit_values takes no arguments
        self.re_emit_values()

    def _get_aperture_value(self):
        """Get the size and the label of the aperture in place.
        Returns:
            (list, str): Size [mm] [width, height], label.
        """
        _size = self.aperture.get_value().value[1]
        try:
            _label = self.aperture.get_value().name
        except AttributeError:
            _label = str(_size)
        _size /= 1000.0

        return [_size, _size], _label

    def _get_definer_value(self):
        """Get the size and the name of the definer in place.
        Returns:
            (list, str): Size [mm] [width, height], label.
        """
        try:
            value = self.definer.get_value()
            if isinstance(value, tuple):
                return [value[1], value[1]], value[0]
            if value.name != "UNKNOWN":
                return list(value.value), value.name
        except AttributeError:
            logging.getLogger("HWR").info("Could not read beam size")
        return [-1, -1], "UNKNOWN"

    def _get_slits_size(self):
        """Get the size of the slits in place.
        Returns:
            (dict): {"width": float, "heigth": float}.
        """
        beam_size = {}
        for _key, _val in self.slits:
            beam_size.update({_key: abs(_val.position)})
        return beam_size

    def get_value(self):
        """Get the size (width and heigth) of the beam, its shape and
           its label. The size is in mm.
        Retunrs:
            (tuple): (width, heigth, shape, name), with types
                     (float, float, Enum, str)
        """
        labels = {}
        _label = "UNKNOWN"
        if self.aperture:
            _size, _name = self._get_aperture_value()
            self._beam_size_dict.update({"aperture": _size})
            labels.update({"aperture": _name})

        if self.slits:
            _size, _name = self._get_slits_value()
            self._beam_size_dict.update({"slits": _size})
            labels.update({"slits": _name})

        if self.definer:
            _size, _name = self._get_definer_value()
            self._beam_size_dict.update({"definer": _size})
            labels.update({"definer": _name})

        info_dict = self.evaluate_beam_info()

        try:
            _label = labels[info_dict["label"]]
            self._beam_info_dict["label"] = _label
        except KeyError:
            _label = info_dict["label"]

        return self._beam_width, self._beam_height, self._beam_shape, _label

    def get_value_xml(self):
        """XMLRPC does not handle Enum, the shape is transformed to string"""
        beamsize = self.get_value()
        return beamsize[0], beamsize[1], beamsize[2].value, beamsize[3]

    def get_available_size(self):
        """Get the available predefined beam definer configuration.
        Returns:
            (dict): {"type": ["apertures"], "values": [labels]} or
                    {"type": ["definer"], "values": [labels]} or
                    {"type": ["width", "height"], "values":
                             [low_lim_w, high_lim_w, low_lim_h, high_lim_h]}
        """
        if self._definer_type == "aperture":
            return {
                "type": ["aperture"],
                "values": self.aperture.get_diameter_size_list(),
            }

        if self._definer_type == "definer":
            return {
                "type": ["definer"],
                "values": self.definer.get_predefined_positions_list(),
            }

        if self._definer_type in (self.slits, "slits"):
            # get the list of the slits motors range
            _low_w, _high_w = self.slits["width"].get_limits()
            _low_h, _high_h = self.slits["height"].get_limits()
            return {
                "type": ["width", "height"],
                "values": [_low_w, _high_w, _low_h, _high_h],
            }

        return {}

    def get_defined_beam_size(self):
        """Get the predefined beam labels and size.
        Returns:
            (dict): Dictionary wiith list of avaiable beam size labels
                    and the corresponding size (width,height) tuples.
                    {"label": [str, str, ...], "size": [(w,h), (w,h), ...]}
        """
        labels = []
        values = []

        if self._definer_type == "slits":
            # get the list of the slits motors range
            _low_w, _high_w = self.slits["width"].get_limits()
            _low_h, _high_h = self.slits["height"].get_limits()
            return {
                "label": ["low", "high"],
                "size": [(_low_w, _low_h), (_high_w, _high_h)],
            }

        if self._definer_type == "aperture":
            _enum = self.aperture.VALUES
        elif self._definer_type == "definer":
            _enum = self.definer.VALUES

        for value in _enum:
            _nam = value.name
            if _nam not in ["IN", "OUT", "UNKNOWN"]:
                labels.append(_nam)
                if isinstance(value.value, tuple):
                    values.append(value.value)
                else:
                    values.append(value.value[0])
        return {"label": labels, "size": values}

    def _set_slits_size(self, size=None):
        """Move the slits to the desired position.
        Args:
            size (list): Width, heigth [mm].
        Raises:
            RuntimeError: Size out of the limits.
               TypeError: Invalid size
        """
        if not isinstance(size, list):
            raise TypeError("Incorrect input value for slits")
        w_lim = self.slits["width"].get_limits()
        h_lim = self.slits["heigth"].get_limits()
        try:
            if min(w_lim) > size[0] > max(w_lim):
                raise RuntimeError("Size out of the limits")
            if min(h_lim) > size[1] > max(h_lim):
                raise RuntimeError("Size out of the limits")
            self.slits["width"].set_value(size[0])
            self.slits["heigth"].set_value(size[1])
        except TypeError as err:
            raise TypeError("Invalid size") from err

    def _set_aperture_size(self, size=None):
        """Move the aperture to the desired size.
        Args:
            size (str): The position name.
        Raises:
            TypeError
        """
        if not isinstance(size, str):
            raise TypeError("Incorrect input value for aperture")

        try:
            _ap = self.aperture.VALUES[size]
        except KeyError:
            _ap = self.aperture.VALUES[f"A{size}"]

        self.aperture.set_value(_ap)

    def _set_definer_size(self, size=None):
        """Move the definer to the desired size.
        Args:
            size (str): The position name.
        Raises:
            TypeError: Invalid size.
        """
        if not isinstance(size, str):
            raise TypeError("Incorrect input value for definer")

        self._definer.set_value(self.definer.VALUES[size])

    def set_value(self, size=None):
        """Set the beam size
        Args:
            size (list): Width, heigth or
                  (str): Aperture or definer name.
        Raises:
            RuntimeError: Beam definer not configured
                          Size out of the limits.
        """
        if self._definer_type in (self.slits, "slits"):
            self._set_slits_size(size)

        if self._definer_type in (self.aperture, "aperture"):
            self._set_aperture_size(size)

        if self._definer_type in (self.definer, "definer"):
            self._set_definer_size(size)

    def get_beam_position_on_screen(self):
        if self._beam_position_on_screen == [0, 0]:
            try:
                _beam_position_on_screen = (
                    HWR.beamline.diffractometer.get_beam_position()
                )
            except AttributeError:
                _beam_position_on_screen = (
                    HWR.beamline.sample_view.camera.get_width() / 2,
                    HWR.beamline.sample_view.camera.get_height() / 2,
                )
            self._beam_position_on_screen = _beam_position_on_screen
        return self._beam_position_on_screen

    def get_beam_size(self):
        beam_value = self.get_value()
        return (beam_value[0], beam_value[1])



#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

from mxcubecore import HardwareRepository as HWR
from mxcubecore.model import queue_model_objects
from mxcubecore.queue_entry.base_queue_entry import BaseQueueEntry

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class XrayCenteringQueueEntry(BaseQueueEntry):
    """
    Defines the behaviour of an Advanced scan
    """

    def __init__(self, view=None, data_model=None, view_set_queue_entry=True):

        BaseQueueEntry.__init__(self, view, data_model, view_set_queue_entry)
        self.mesh_qe = None
        self.helical_qe = None
        self.in_queue = False

    def execute(self):
        BaseQueueEntry.execute(self)

    def pre_execute(self):
        BaseQueueEntry.pre_execute(self)
        xray_centering = self.get_data_model()
        reference_image_collection = xray_centering.reference_image_collection
        reference_image_collection.grid = HWR.beamline.sample_view.create_auto_grid()
        reference_image_collection.acquisitions[
            0
        ].acquisition_parameters.centred_position = (
            reference_image_collection.grid.get_centred_position()
        )

        # Trick to make sure that the reference collection has a sample.
        reference_image_collection._parent = xray_centering.get_parent()
        xray_centering.line_collection._parent = xray_centering.get_parent()

        gid = self.get_data_model().get_parent().lims_group_id
        reference_image_collection.lims_group_id = gid

        # Enqueue the reference mesh scan collection
        mesh_qe = DataCollectionQueueEntry(
            self.get_view(), reference_image_collection, view_set_queue_entry=False
        )
        mesh_qe.set_enabled(True)
        mesh_qe.in_queue = self.in_queue
        self.mesh_qe = mesh_qe

        # Creat e a helical data collection based on the first collection
        helical_qe = DataCollectionQueueEntry(
            self.get_view(), reference_image_collection, view_set_queue_entry=False
        )

        # helical_model = helical_qe.get_data_model()
        # @helical_model.set_experiment_type(EXPERIMENT_TYPE.HELICAL)
        # @helical_model.grid = None

        acq_two = queue_model_objects.Acquisition()
        helical_model.acquisitions.append(acq_two)
        helical_model.acquisitions[0].acquisition_parameters.num_images = 100
        helical_model.acquisitions[0].acquisition_parameters.num_lines = 1
        helical_acq_path_template = helical_model.acquisitions[0].path_template
        helical_acq_path_template.base_prefix = (
            "line_" + helical_acq_path_template.base_prefix
        )
        helical_qe._data_model = helical_model

        helical_qe.set_enabled(True)
        helical_qe.in_queue = self.in_queue
        self.helical_qe = helical_qe

        advanced_connector_qe = AdvancedConnectorQueueEntry(
            self.get_view(), reference_image_collection, view_set_queue_entry=False
        )
        advanced_connector_qe.first_qe = mesh_qe
        advanced_connector_qe.second_qe = helical_qe
        advanced_connector_qe.set_enabled(True)

        self.enqueue(mesh_qe)
        self.enqueue(advanced_connector_qe)
        self.enqueue(helical_qe)

    def post_execute(self):
        if self.helical_qe:
            self.status = self.helical_qe.status
        else:
            self.status = self.mesh_qe
        BaseQueueEntry.post_execute(self)



from mxcubecore.BaseHardwareObjects import HardwareObject

from .BeamCmds import (
    ControllerCommand,
    HWObjActuatorCommand,
    TestCommand,
)


class ID30BBeamCmds(HardwareObject):
    """Beam action commands for ID30-A3"""
    def __init__(self, *args):
        HardwareObject.__init__(self, *args)

    def init(self):
        controller = self.get_object_by_role("controller")
        controller.detcover.set_in()
        self.centrebeam = ControllerCommand("Centre beam", controller.centrebeam)
        hutchtrigger = self.get_object_by_role("hutchtrigger")
        if hutchtrigger:
            self.hutchtrigger = HWObjActuatorCommand("Hutchtrigger", hutchtrigger)

        detector_cover = self.get_object_by_role("detector_cover")
        if detector_cover:
            self.detcover = HWObjActuatorCommand("Detector Cover", detector_cover)
        self.test_cmd = TestCommand("Test command")
        # self.quick_realign = ControllerCommand(
        #   "Quick realign", controller.quick_realign
        # )
        # self.anneal = ControllerCommand("Anneal", controller.anneal_procedure)

    def get_commands(self):
        return [self.centrebeam] #self.hutchtrigger]#, self.detcover]



import logging

from beam_center import beam_center
from resolution import resolution

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractResolution import AbstractResolution


class PX2Resolution(AbstractResolution):
    def __init__(self, name):
        super(PX2Resolution, self).__init__(name)
        self.resolution_motor = resolution()
        self.beam_center = beam_center()
        
    def connect_notify(self, signal):
        if signal == "stateChanged":
            self.update_state(self.get_state())

    def get_value(self):
        self._nominal_value = self.resolution_motor.get_resolution()
        return self._nominal_value
    
    def _set_value(self, value):
        self.resolution_motor.set_resolution(value)

    def get_beam_centre(self, dtox=None):
        return self.beam_center.get_beam_center()

    def get_limits(self):
        return self.resolution_motor.get_resolution_limits()

    def stop(self):
        self.resolution_motor.stop()

    def is_ready(self):
        return True

    def update_distance(self, value=None):
        """Update the resolution when distance changed.
        Args:
            value (float): Detector distance [mm].
        """
        self._nominal_value = self.resolution_motor.get_resolution()
        self.emit("valueChanged", (self._nominal_value,))



from mxcubecore import BaseHardwareObjects
from mxcubecore import HardwareRepository as HWR


class Q315dist(BaseHardwareObjects.Equipment):
    def _init(self):
        self.connect("equipmentReady", self.equipmentReady)
        self.connect("equipmentNotReady", self.equipmentNotReady)

        return BaseHardwareObjects.super()._init()

    def init(self):
        self.detm = self.get_deviceby_role("detm")

        self.connect(self.detm, "stateChanged", self.detmStateChanged)
        self.connect(
            HWR.beamline.detector.distance, "limitsChanged", self.dtoxLimitsChanged
        )
        self.connect(self.detm, "valueChanged", self.detmPositionChanged)

    def equipmentReady(self):
        self.emit("deviceReady")

    def equipmentNotReady(self):
        self.emit("deviceNotReady")

    def is_valid(self):
        return (
            self.get_deviceby_role("detm") is not None
            and self.get_deviceby_role("detector_distance") is not None
        )

    def __getattr__(self, attr):
        """Delegation to underlying motors"""
        if not attr.startswith("__"):
            if attr in ("get_value", "get_state", "get_limits"):
                # logging.getLogger().info("calling detm %s ; ready ? %s", attr, self.detm.is_ready())
                return getattr(self.detm, attr)
            else:
                # logging.getLogger().info("calling dtox %s", attr)
                return getattr(HWR.beamline.detector.distance, attr)
        else:
            raise AttributeError(attr)

    def connect_notify(self, signal):
        if signal == "stateChanged":
            self.detmStateChanged(self.detm.get_state())
        elif signal == "valueChanged":
            self.detmPositionChanged(self.detm.get_value())

    def detmStateChanged(self, state):
        if (state == self.detm.NOTINITIALIZED) or (state > self.detm.UNUSABLE):
            self.emit("stateChanged", (state,))
        else:
            self.detm.motorState = self.detm.READY
            self.detm.motorStateChanged(self.detm.motorState)

    def dtoxLimitsChanged(self, limits):
        self.emit("limitsChanged", (limits,))

    def detmPositionChanged(self, pos):
        self.emit("valueChanged", (pos,))



import logging

import taurus

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract import AbstractFlux


class ALBAFlux(Device, AbstractFlux.AbstractFlux):
    def init(self):
        super(ALBAFlux, self).init()
        self.current_dev = taurus.Device("pc/mach_attrpc_ctrl/1")
        self.vars_dev = taurus.Device("bl13/ct/variables")
        self.trans_mot = taurus.Device("mbattrans")

    def get_value(self):
        fluxlast = self.vars_dev["fluxlast"].value

        try:
            if fluxlast > 1e7:
                return self.last_current_trans()
        except Exception:
            pass

        logging.getLogger("HWR").debug(
            " Abnormally low value of flux. Returning default value"
        )
        default_flux = 6e11 * self.get_transmission()
        return default_flux

    def get_transmission(self):
        """ returns transmission between 0 and 1"""
        return self.trans_mot.position / 100.0

    def last_current_trans(self):
        current = self.current_dev.value
        fluxlastnorm = self.vars_dev["fluxlastnorm"].value

        last_current = (fluxlastnorm / 250.0) * current

        return last_current * self.get_transmission()


def test_hwo(hwo):
    print(hwo.get_value())



import logging

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject


class ALBAFrontLight(HardwareObject):
    def __init__(self, *args):
        super().__init__(*args)
        self.limits = [None, None]

        self.state = None
        self.register_state = None

        self.current_level = None
        self.memorized_level = None
        self.previous_level = None

        self.default_off_threshold = 0.01
        self.off_threshold = None

    def init(self):

        self.level_channel = self.get_channel_object("light_level")
        self.state_channel = self.get_channel_object("state")
        threshold = self.get_property("off_threshold")

        if threshold is not None:
            try:
                self.off_threshold = float(threshold)
            except Exception:
                self.off_threshold = self.default_threshold
                logging.getLogger("HWR").info(
                    "OFF Threshold for front light is not valid. Using %s"
                    % self.off_threshold
                )

        limits = self.get_property("limits")
        if limits is not None:
            lims = limits.split(",")
            if len(lims) == 2:
                self.limits = map(float, lims)

        self.level_channel.connect_signal("update", self.level_changed)
        self.state_channel.connect_signal("update", self.register_state_changed)

    def is_ready(self):
        return True

    def level_changed(self, value):
        self.current_level = value
        self.update_current_state()

        self.emit("levelChanged", self.current_level)

    def register_state_changed(self, value):
        self.register_state = str(value).lower()
        self.update_current_state()

    def update_current_state(self):
        if self.register_state == "on":
            if (
                self.off_threshold is not None
                and self.current_level < 0.9 * self.off_threshold
            ):
                newstate = "off"
            else:
                newstate = "on"
        elif self.register_state == "off":
            newstate = "off"
        else:
            newstate = "fault"

        if newstate != self.state:
            if newstate == "off":
                self.memorized_level = self.previous_level

        self.state = newstate
        self.emit("stateChanged", self.state)

        self.previous_level = self.current_level

    def get_limits(self):
        return self.limits

    def get_state(self):
        self.register_state = str(self.state_channel.get_value()).lower()
        self.update_current_state()
        return self.state

    def getUserName(self):
        return self.username

    def getLevel(self):
        self.current_level = self.level_channel.get_value()
        return self.current_level

    def setLevel(self, level):
        logging.getLogger("HWR").debug(
            "Setting level in %s to %s" % (self.username, level)
        )
        self.level_channel.set_value(float(level))

    def setOn(self):
        logging.getLogger("HWR").debug("Setting front light on")
        if self.memorized_level is not None:
            if self.memorized_level < self.off_threshold:
                value = self.off_threshold
            else:
                value = self.memorized_level
            logging.getLogger("HWR").debug("   setting value to")
            self.level_channel.set_value(value)
        else:
            self.level_channel.set_value(self.off_threshold)

    def setOff(self):
        logging.getLogger("HWR").debug("Setting front light off")
        self.level_channel.set_value(0.0)


def test():
    hwr = HWR.get_hardware_repository()
    hwr.connect()

    light = hwr.get_hardware_object("/frontlight")
    print('\nLight control for "%s"\n' % light.getUserName())
    print("   Level limits are:", light.get_limits())
    print("   Current level is:", light.getLevel())
    print("   Current state is:", light.get_state())


if __name__ == "__main__":
    test()



#! /usr/bin/env python
# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Beamline class serving as singleton container for links to top-level HardwareObjects

All HardwareObjects
"""

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

from typing import (
    Any,
    Union,
)

from mxcubecore.dispatcher import dispatcher

__copyright__ = """ Copyright © 2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"
__author__ = "Rasmus H Fogh"

import logging

from mxcubecore.BaseHardwareObjects import (
    ConfiguredObject,
    HardwareObject,
)

# NBNB The acq parameter names match the attributes of AcquisitionParameters
# Whereas the limit parameter values use more understandable names
#
# TODO Make all tags consistent, including AcquisitionParameters attributes.


class Beamline(ConfiguredObject):
    """Beamline class serving as singleton container for links to HardwareObjects"""

    # Roles of defined objects and the category they belong to
    # NB the double underscore is deliberate - attribute must be hidden from subclasses
    __content_roles = []

    # Names of procedures under Beamline - set of strings.
    # NB subclasses must add additional procedures to this set,
    # and may NOT override _procedure_names
    _procedure_names = set()

    # NBNB these should be accessed ONLY as beamline.SUPPORTED_..._PARAMETERS
    # NBNB Subclasses may add local parameters but may NOT remove any
    #
    # Supported acquisition parameter tags:
    SUPPORTED_ACQ_PARAMETERS = frozenset(
        (
            "exp_time",
            "osc_range",
            "num_passes",
            "first_image",
            "run_number",
            "overlap",
            "num_images",
            "inverse_beam",
            "take_dark_current",
            "skip_existing_images",
            "take_snapshots",
        )
    )
    # Supported limit parameter tags:
    SUPPORTED_LIMIT_PARAMETERS = frozenset(
        ("exposure_time", "osc_range", "number_of_images", "kappa", "kappa_phi")
    )

    def __init__(self, name):
        """

        Args:
            name (str) : Object name, generally set to the role name of the object
        """
        super(Beamline, self).__init__(name)

        # List[str] of advanced method names
        self.advanced_methods = []

        # List[str] of available methods
        self.available_methods = []

        # int number of clicks used for click centring
        self.click_centring_num_clicks = 3

        # bool Is wavelength tunable
        self.tunable_wavelength = False

        # bool Disable number-of-passes widget NBNB TODO Move elsewhere??
        self.disable_num_passes = False

        # bool By default run processing of (certain?)data collections?
        self.run_offline_processing = False

        # bool By default run online processing (characterization/mesh?)
        self.run_online_processing = False

        self.offline_processing_methods = []

        self.online_processing_methods = []

        # Dictionary-of-dictionaries of default acquisition parameters
        self.default_acquisition_parameters = {}

        # Dictionary of acquisition parameter limits
        self.acquisition_limit_values = {}

        # int Starting run number for path_template
        self.run_number = 1

        # List of undulators
        self.undulators = []

        # Format of mesh result for display
        self.mesh_result_format = "PNG"

        # bool Use the native mesh feature available, true by default
        self.use_native_mesh = True

        # bool Enable features to work with points in the plane, called
        # 2D-points, (none centred positions)
        self.enable_2d_points = True

        # Dictionary with the python id of hardwareobject as key
        # and the "dotted/attribute path" to hardwareobject from the
        # Beamline object
        self._hardware_object_id_dict = {}

    def init(self):
        """Object initialisation - executed *after* loading contents"""
        # Validate acquisition parameters
        for acquisition_type, params in self.default_acquisition_parameters.items():
            unrecognised = [x for x in params if x not in self.SUPPORTED_ACQ_PARAMETERS]
            if unrecognised:
                logging.getLogger("HWR").warning(
                    "Unrecognised acquisition parameters for %s: %s"
                    % (acquisition_type, unrecognised)
                )
        # Validate limits parameters
        unrecognised = [
            x
            for x in self.acquisition_limit_values
            if x not in self.SUPPORTED_LIMIT_PARAMETERS
        ]
        if unrecognised:
            logging.getLogger("HWR").warning(
                "Unrecognised parameter limits for: %s" % unrecognised
            )

    def _hwr_init_done(self):
        """
        Method called after the initialization of HardwareRepository is done
        (when all HardwareObjects have been created and initialized)
        """
        self._hardware_object_id_dict = self._get_id_dict()

    def get_id(self, ho: HardwareObject) -> str:
        """
        Returns "dotted path/attribute" which is unique within the context of
        HardwareRepository

        Args:
            ho: The hardware object for which to get the id

        Returns:
            "dotted path/attribute"
        """
        return self._hardware_object_id_dict.get(ho)

    def get_hardware_object(self, _id: str) -> Union[HardwareObject, None]:
        """
        Returns the HardwareObject with the given id

        Args:
            _id: "attribute path" / id of HardwareObject
        Returns:
            HardwareObject with the given id
        """
        found_ho = None

        for current_ho, current_id in self._hardware_object_id_dict.items():
            if current_id == _id:
                found_ho = current_ho

        return found_ho

    def _get_id_dict(self) -> dict:
        """
        Wrapper function used to call the recursive method used to find all
        HardwareObjects accessible from the Beamline object.
        """
        result = {}

        for ho_name in self.all_roles:
            ho = self._objects.get(ho_name)

            if ho:
                result[ho] = ho_name
                self._get_id_dict_rec(ho, ho_name, result)

        return result

    def _get_id_dict_rec(
        self, ho: HardwareObject, _path: str = "", result: dict = {}
    ) -> str:
        """
        Recurses through all the roles of ho and constructs its corresponding
        "dotted path/attribute"

        Args:
            ho (HardwareObject): The HardwareObject to get the id for
            _path (str): Current path (used in recursion)
            result: A dictionary where the key is the id of the HardwareObject
                    and the value its dotted path.

        Returns:
            (str): Dotted path for the given HardwareObject
        """
        if hasattr(ho, "get_roles"):
            for role in ho.get_roles():
                child_ho = ho.get_object_by_role(role)
                if child_ho not in result:
                    result[child_ho] = self._get_id_dict_rec(
                        child_ho, f"{_path}.{role}", result
                    )

        return _path

    # Signal handling functions:
    def emit(self, signal: Union[str, object, Any], *args) -> None:
        """Emit signal. Accepts both multiple args and a single tuple of args.

        This is needed for communication from the GUI to the core
        (jsonparamsgui in mxcubeqt)

        NBNB TODO HACK
        This is a duplicate of the same function in HardwareObjectMixin.
        Since the Beamline is not a CommandContainer or a normal HardwareObject
        it may not be appropriate to make it a subclass of HardwareObjectYaml
        We need to consider how we want this organised

        Args:
            signal (Union[str, object, Any]): In practice a string, or dispatcher.
            *args (tuple): Arguments sent with signal.
        """

        signal = str(signal)

        if len(args) == 1:
            if isinstance(args[0], tuple):
                args = args[0]
        responses: list = dispatcher.send(signal, self, *args)
        if not responses:
            raise RuntimeError(
                "Signal %s is not connected" % signal
            )

    # NB this function must be re-implemented in nested subclasses
    @property
    def all_roles(self):
        """Tuple of all content object roles, indefinition and loading order

        Returns:
            tuple[text_str, ...]
        """
        return super(Beamline, self).all_roles + tuple(self.__content_roles)

    @property
    def machine_info(self):
        """Machine information Hardware object

        Returns:
            Optional[AbstractMachineInfo]:
        """
        return self._objects.get("machine_info")

    __content_roles.append("machine_info")

    @property
    def authenticator(self):
        """Authenticator Hardware object

        Returns:
            Optional[AbstractAuthenticator]:
        """
        return self._objects.get("authenticator")

    __content_roles.append("authenticator")

    @property
    def transmission(self):
        """Transmission Hardware object

        Returns:
            Optional[AbstractTransmission]:
        """
        return self._objects.get("transmission")

    __content_roles.append("transmission")

    @property
    def cryo(self):
        """Cryo Hardware object

        Returns:
            Optional[AbstractActuator]:
        """
        return self._objects.get("cryo")

    __content_roles.append("cryo")

    @property
    def energy(self):
        """Energy Hardware object

        Returns:
            Optional[AbstractEnergy]:
        """
        return self._objects.get("energy")

    __content_roles.append("energy")

    @property
    def flux(self):
        """Flux Hardware object

        Returns:
            Optional[AbstractActuator]:
        """
        return self._objects.get("flux")

    __content_roles.append("flux")

    @property
    def beam(self):
        """Beam Hardware object

        Returns:
            Optional[AbstractBeam]:
        """
        return self._objects.get("beam")

    __content_roles.append("beam")

    @property
    def hutch_interlock(self):
        """Hutch Interlock Hardware object

        Returns:
            Optional[AbstractInterlock]:
        """
        return self._objects.get("hutch_interlock")

    __content_roles.append("hutch_interlock")

    @property
    def sample_environment(self):
        """Sample Environment Hardware Object

        Returns:
            Optional[AbstractSampleEnvironment]:
        """
        return self._objects.get("sample_environment")

    __content_roles.append("sample_environment")

    @property
    def safety_shutter(self):
        """Safety Shutter Hardware object

        Returns:
            Optional[AbstractShutter]:
        """
        return self._objects.get("safety_shutter")

    __content_roles.append("safety_shutter")

    @property
    def fast_shutter(self):
        """Fast Shutter Hardware object

        Returns:
            Optional[AbstractShutter]:
        """
        return self._objects.get("fast_shutter")

    __content_roles.append("fast_shutter")

    @property
    def diffractometer(self):
        """Diffractometer Hardware object

        Returns:
            Optional[AbstractDiffractometer]:
        """
        return self._objects.get("diffractometer")

    __content_roles.append("diffractometer")

    @property
    def detector(self):
        """Detector Hardware object

        Returns:
            Optional[AbstractDetector]:
        """
        return self._objects.get("detector")

    __content_roles.append("detector")

    @property
    def resolution(self):
        """Resolution Hardware object

        Returns:
            Optional[AbstractActuator]:
        """
        return self._objects.get("resolution")

    __content_roles.append("resolution")

    @property
    def sample_changer(self):
        """Sample Changer Hardware object
        can be a sample changer, plate_manipulator, jets, chips

        Returns:
            Optional[AbstractSampleChanger]:
        """
        return self._objects.get("sample_changer")

    __content_roles.append("sample_changer")

    @property
    def sample_changer_maintenance(self):
        """Sample Changer Maintnance Hardware object

        Returns:
            Optional[AbstractMaintnanceSampleChanger]:
        """
        return self._objects.get("sample_changer_maintenance")

    __content_roles.append("sample_changer_maintenance")

    @property
    def harvester(self):
        """Harvester Hardware object
        can be a sample or plate holder

        Returns:
            Optional[AbstractHarvester]:
        """
        return self._objects.get("harvester")

    __content_roles.append("harvester")

    @property
    def harvester_maintenance(self):
        """harvester maintenance Hardware object

        Returns:
            Optional[Harvester]:
        """
        return self._objects.get("harvester_maintenance")

    __content_roles.append("harvester_maintenance")

    @property
    def plate_manipulator(self):
        """**DEPRECATED**
        Plate Manipulator Hardware object
        NBNB TODO REMOVE THIS From qt version usage and
        and call HWR.beamline.sample_changer instead as plate_manipulator being
        treated as an alternative sample changer.

        Returns:
            Optional[AbstractSampleChanger]:
        """
        return self._objects.get("plate_manipulator")

    __content_roles.append("plate_manipulator")

    @property
    def session(self):
        """Session Hardware object, holding information on current session and user.

        Returns:
            Optional[Session]:
        """
        return self._objects.get("session")

    __content_roles.append("session")

    @property
    def lims(self):
        """LIMS client object.

        Returns:
            Optional[ISPyBClient]:
        """
        return self._objects.get("lims")

    __content_roles.append("lims")

    @property
    def sample_view(self):
        """Sample view object. Includes defined shapes.

        Returns:
            Optional[AbstractSampleView]:
        """
        return self._objects.get("sample_view")

    __content_roles.append("sample_view")

    @property
    def queue_manager(self):
        """Queue manager object.

        Returns:
            Optional[QueueManager]:
        """
        return self._objects.get("queue_manager")

    __content_roles.append("queue_manager")

    @property
    def queue_model(self):
        """Queue model object.

        Returns:
            Optional[QueueModel]:
        """
        return self._objects.get("queue_model")

    __content_roles.append("queue_model")

    # Procedures

    @property
    def collect(self):
        """Data collection procedure.

        Returns:
            Optional[AbstractCollect]:
        """
        return self._objects.get("collect")

    __content_roles.append("collect")

    @property
    def xrf_spectrum(self):
        """X-ray fluorescence spectrum procedure.

        Returns:
            Optional[AbstractProcedure]
        """
        return self._objects.get("xrf_spectrum")

    __content_roles.append("xrf_spectrum")

    @property
    def energy_scan(self):
        """Energy scan procedure.

        Returns:
            Optional[AbstractProcedure]:
        """
        return self._objects.get("energy_scan")

    __content_roles.append("energy_scan")

    @property
    def imaging(self):
        """Imaging procedure.

        Returns:
            Optional[AbstractProcedure]:
        """
        return self._objects.get("imaging")

    __content_roles.append("imaging")

    @property
    def beamline_actions(self):
        """Beamline Actions

        Returns:
            Optional[beamline_actions]:
        """
        return self._objects.get("beamline_actions")

    __content_roles.append("beamline_actions")

    @property
    def xml_rpc_server(self):
        """XMLRPCServer for RPC

        Returns:
            Optional[XMLRPCServer]:
        """
        return self._objects.get("xml_rpc_server")

    __content_roles.append("xml_rpc_server")

    @property
    def workflow(self):
        """Standarad EDNA workflow procedure.

        Returns:
            Optional[Workflow]:
        """
        return self._objects.get("workflow")

    __content_roles.append("workflow")

    @property
    def control(self):
        """Beamline control system

        Returns:
            Optional[Control]:
        """
        return self._objects.get("control")

    __content_roles.append("control")

    @property
    def gphl_workflow(self):
        """Global phasing data collection workflow procedure.

        Returns:
            Optional[GphlWorkflow]:
        """
        return self._objects.get("gphl_workflow")

    __content_roles.append("gphl_workflow")

    # This one is 'hardware', but it is put with its companion
    @property
    def gphl_connection(self):
        """Global PHasing workflow remote connection

        Returns:
            Optional[GphlWorkflowConnection]:
        """
        return self._objects.get("gphl_connection")

    __content_roles.append("gphl_connection")

    # centring

    # NB Could centring be treated as procedures instesad?

    @property
    def centring(self):
        """Centring procedures object. Includes X-ray, n-click, optical, move_to_beam

        Returns:
            Optional[AbstractCentring]:
        """
        return self._objects.get("centring")

    __content_roles.append("centring")

    @property
    def xray_centring(self):
        """Xray Ccntring hardware object.

        Returns:
            Optional[XrayCentring2]:
        """
        return self._objects.get("xray_centring")

    __content_roles.append("xray_centring")

    # Analysis (combines processing and data analysis)

    @property
    def online_processing(self):
        """Synchronous (on-line) data processing procedure.

        Returns:
            Optional[AbstractProcessing]:
        """
        return self._objects.get("online_processing")

    __content_roles.append("online_processing")

    @property
    def offline_processing(self):
        """Asynchronous (queue sumbission) data processing procedure.

        Returns:
            Optional[AbstractProcessing]:
        """
        return self._objects.get("offline_processing")

    __content_roles.append("offline_processing")

    @property
    def characterisation(self):
        """EDNA characterisation and analysis procedure.

        NB the current code looks rather EDNA-specific
        to be called 'AbsatractCharacterisation'.
        Potentially we could generalise it, and maybe make it into a procedure???

        Returns:
            Optional[EdnaCharacterisation]:
        """
        return self._objects.get("characterisation")

    __content_roles.append("characterisation")

    @property
    def beam_realign(self):
        """Beam-realign procedure object

        Returns:
            Optional[AbstractProcedure]:
        """
        return self._objects.get("beam_realign")

    __content_roles.append("beam_realign")

    @property
    def image_tracking(self):
        """Imaging tracking object

        Returns:
            Optional[HardwareObject]:
        """
        return self._objects.get("image_tracking")

    __content_roles.append("image_tracking")

    # Procedures

    @property
    def mock_procedure(self):
        """ """
        return self._objects.get("mock_procedure")

    __content_roles.append("mock_procedure")

    @property
    def data_publisher(self):
        """ """
        return self._objects.get("data_publisher")

    __content_roles.append("data_publisher")

    # NB this is just an example of a globally shared procedure description
    @property
    def manual_centring(self):
        """Manual centring Procedure

        NB AbstractManualCentring serves to define the parameters for manual centring
        The actual implementation is set by configuration,
        and can be given as an AbstractManualCentring subclass on each beamline

        Returns:
            Optional[AbstractManualCentring]
        """
        return self._objects.get("manual_centring")

    __content_roles.append("manual_centring")
    # Registers this object as a procedure:
    _procedure_names.add("manual_centring")

    # Additional functions

    # NB Objects need not be HardwareObjects
    # We still categorise them as'hardware' if they are not procedures, though
    # The attribute values will be given in the config.yml file
    def get_default_acquisition_parameters(self, acquisition_type="default"):
        """
        :returns: A AcquisitionParameters object with all default parameters for the
                  specified acquisition type. "default" is a standard acqquisition
        """
        # Imported here to avoid circular imports
        from mxcubecore.model import queue_model_objects

        acq_parameters = queue_model_objects.AcquisitionParameters()

        params = self.default_acquisition_parameters["default"].copy()
        if acquisition_type != "default":
            dd0 = self.default_acquisition_parameters.get(acquisition_type)
            if dd0 is None:
                logging.getLogger("HWR").warning(
                    "No separate parameters for acquisition type: %s - using default."
                    % acquisition_type
                )
            else:

                params.update(dd0)

        for tag, val in params.items():
            setattr(acq_parameters, tag, val)

        motor_positions = self.diffractometer.get_positions()
        osc_start = motor_positions.get("phi")
        if osc_start is None:
            acq_parameters.osc_start = params.get("osc_start")
        else:
            acq_parameters.osc_start = osc_start

        kappa = motor_positions.get("kappa")
        if kappa is None:
            acq_parameters.kappa = None
        else:
            acq_parameters.kappa = round(float(kappa), 2)

        kappa_phi = motor_positions.get("kappa_phi")
        if kappa_phi is None:
            acq_parameters.kappa_phi = None
        else:
            acq_parameters.kappa_phi = round(float(kappa_phi), 2)

        try:
            acq_parameters.resolution = self.resolution.get_value()
        except Exception:
            logging.getLogger("HWR").warning(
                "get_default_acquisition_parameters: "
                "No current resolution, setting to 0.0"
            )
            acq_parameters.resolution = 0.0

        try:
            acq_parameters.energy = self.energy.get_value()
        except Exception:
            logging.getLogger("HWR").warning(
                "get_default_acquisition_parameters: "
                "No current energy, setting to 0.0"
            )
            acq_parameters.energy = 0.0

        try:
            acq_parameters.transmission = self.transmission.get_value()
        except Exception:
            logging.getLogger("HWR").warning(
                "get_default_acquisition_parameters: "
                "No current transmission, setting to 0.0"
            )
            acq_parameters.transmission = 0.0

        acq_parameters.shutterless = params.get("shutterless", True)

        try:
            acq_parameters.detector_binning_mode = self.detector.get_binning_mode()
        except Exception:
            logging.getLogger("HWR").warning(
                "get_default_acquisition_parameters: "
                "Could not get detector mode, setting to ''"
            )
            acq_parameters.detector_binning_mode = ""

        try:
            acq_parameters.detector_roi_mode = self.detector.get_roi_mode()
        except Exception:
            logging.getLogger("HWR").warning(
                "get_default_acquisition_parameters: "
                "Could not get roi mode, setting to ''"
            )
            acq_parameters.detector_roi_mode = ""

        return acq_parameters

    def get_default_path_template(self):
        """
        :returns: A PathTemplate object with default parameters.
        """
        # Imported here to avoid circular imports
        from mxcubecore.model import queue_model_objects

        path_template = queue_model_objects.PathTemplate()

        acq_params = self.get_default_acquisition_parameters()
        path_template.start_num = acq_params.first_image
        path_template.num_files = acq_params.num_images

        path_template.run_number = self.run_number

        return path_template

    def get_default_characterisation_parameters(self):
        return self.characterisation.get_default_characterisation_parameters()

    def force_emit_signals(self):
        for role in self.all_roles:
            hwobj = getattr(self, role)
            if hwobj is not None:
                try:
                    hwobj.force_emit_signals()
                    for attr in dir(hwobj):
                        if not attr.startswith("_"):
                            if hasattr(getattr(hwobj, attr), "force_emit_signals"):
                                child_hwobj = getattr(hwobj, attr)
                                child_hwobj.force_emit_signals()
                except BaseException as ex:
                    logging.getLogger("HWR").error(
                        "Unable to call force_emit_signals (%s)" % str(ex)
                    )



"""
[Name] BeamInfo

[Description]
BeamInfo hardware object informs mxCuBE (HutchMenuBrick) about the beam position
and size.

This is the Soleil PX1 version

Beam size is hardcoded in this file.
Beam Position is updated whenever the zoom motor changes position. Values taken from
   zoom xml configuration

[Emited signals]

beamInfoChanged
beamPosChanged

"""

import logging

from mxcubecore.BaseHardwareObjects import HardwareObject


class PX1BeamInfo(HardwareObject):
    def __init__(self, *args):
        super().__init__(*args)

        self.beam_position = [None, None]
        self.beam_size = [100, 100]
        self.shape = "rectangular"

        self.beam_size_slits = [9999, 9999]

        self.beam_info_dict = {}
        self.beam_info_dict["size_x"] = self.beam_size[0]
        self.beam_info_dict["size_y"] = self.beam_size[1]
        self.beam_info_dict["shape"] = self.shape

        # Zoom motor
        self.zoomMotor = None

    def init(self):

        try:
            self.beamx_chan = self.get_channel_object("beamsizex")
        except KeyError:
            logging.getLogger().warning(
                "%s: cannot connect to beamsize x channel ", self.name()
            )

        try:
            self.beamy_chan = self.get_channel_object("beamsizey")
            self.beamy_chan.connect_signal("update", self.beamsize_x_changed)
        except KeyError:
            logging.getLogger().warning(
                "%s: cannot connect to beamsize y channel ", self.name()
            )

        self.zoomMotor = self.get_deviceby_role("zoom")

        if self.beamx_chan is not None:
            self.beamx_chan.connect_signal("update", self.beamsize_x_changed)
        else:
            logging.getLogger().info("BeamSize X channel not defined")

        if self.beamy_chan is not None:
            self.beamy_chan.connect_signal("update", self.beamsize_y_changed)
        else:
            logging.getLogger().info("BeamSize Y channel not defined")

        if self.zoomMotor is not None:
            self.connect(
                self.zoomMotor, "predefinedPositionChanged", self.zoomPositionChanged
            )
            self.zoomPositionChanged()
        else:
            logging.getLogger().info("Zoom motor not defined")

        if None in [self.beamy_chan, self.beamx_chan]:
            try:
                beam_size = self.get_property("beam_size")
                if beam_size is not None:
                    beamx, beamy = beam_size.split(",")
                    self.beam_info_dict["size_x"] = self.beam_size[0] = float(beamx)
                    self.beam_info_dict["size_y"] = self.beam_size[1] = float(beamy)
            except Exception:
                pass

    def connect_notify(self, signal):
        if signal == "beamInfoChanged":
            self.sizeUpdated()
        elif signal == "position_changed":
            self.positionUpdated()

    def zoomPositionChanged(self, name=None, offset=None):
        zoom_props = self.zoomMotor.getCurrentPositionProperties()
        if "beamPositionX" in zoom_props:
            self.beam_position = [
                zoom_props["beamPositionX"],
                zoom_props["beamPositionY"],
            ]
            self.positionUpdated()

    def positionUpdated(self):
        self.emit("beamPosChanged", (self.beam_position,))
        self.sizeUpdated()

    def sizeUpdated(self):
        if None not in [self.beamx_chan, self.beamy_chan]:
            x_beam = self.beamx_chan.get_value()
            y_beam = self.beamy_chan.get_value()
            self.beam_info_dict["size_x"] = x_beam
            self.beam_info_dict["size_y"] = y_beam
        self.emit("beamInfoChanged", (self.beam_info_dict,))

    def get_beam_info(self):
        return self.beam_info_dict

    def get_beam_position(self):
        return self.beam_position

    def get_beam_size(self):
        return self.beam_size

    def get_beam_shape(self):
        return self.shape

    def get_slits_gap(self):
        return self.beam_size_slits

    def get_beam_divergence_hor(self):
        return 0

    def get_beam_divergence_ver(self):
        return 0

    def beamsize_x_changed(self, value=None):
        self.sizeUpdated()

    def beamsize_y_changed(self, value=None):
        self.sizeUpdated()


def test_hwo(hwo):
    print("BEAM info: ", hwo.get_beam_info())
    print("BEAM position: ", hwo.get_beam_position())



import logging
import os
import subprocess


def grouped_processing(processEvent, params):
    endOfLineToExecute = ""

    for param_dict in params:
        dataCollectionId = param_dict.get("collect_id")
        residues = param_dict.get("residues", 0)
        anomalous = param_dict.get("anomalous", False)
        spacegroup = param_dict.get("spacegroup")

        if spacegroup:
            sg_opt = " -sg {0}".format(spacegroup)
        else:
            sg_opt = ""

        unit_cell_constants = param_dict.get("cell")
        if unit_cell_constants:
            cell_opt = ' -cell "{0}"'.format(unit_cell_constants)
        else:
            cell_opt = ""

        endOfLineToExecute += " -mode %s -collect %d:%s" % (
            processEvent,
            dataCollectionId,
            param_dict["xds_dir"],
        )
        endOfLineToExecute += (
            " -residues "
            + str(residues)
            + " -anomalous "
            + str(anomalous)
            + sg_opt
            + cell_opt
        )
        # + (param_dict["inverse_beam"] and ' -inverse' or '')
    return endOfLineToExecute


def start(programs, processEvent, paramsDict):
    for program in programs["program"]:
        try:
            allowed_events = program.get_property("event").split(" ")
            if processEvent in allowed_events:
                executable = program.get_property("executable")

                if os.path.isfile(executable):
                    if processEvent == "end_multicollect":
                        endOfLineToExecute = grouped_processing(
                            "end_multicollect", paramsDict
                        )
                    elif os.path.isdir(paramsDict["xds_dir"]):
                        dataCollectionId = paramsDict.get("datacollect_id")
                        residues = paramsDict.get("residues", 0)
                        anomalous = paramsDict.get("anomalous", False)
                        spacegroup = paramsDict.get("spacegroup")
                        if spacegroup:
                            sg_opt = " -sg {0}".format(spacegroup)
                        else:
                            sg_opt = ""

                        unit_cell_constants = paramsDict.get("cell")
                        if unit_cell_constants:
                            cell_opt = ' -cell "{0}"'.format(unit_cell_constants)
                        else:
                            cell_opt = ""

                        endOfLineToExecute = (
                            " -path "
                            + paramsDict["xds_dir"]
                            + " -mode "
                            + processEvent
                            + " -datacollectionID "
                            + str(dataCollectionId)
                            + " -residues "
                            + str(residues)
                            + " -anomalous "
                            + str(anomalous)
                            + sg_opt
                            + cell_opt
                        )  # +\
                        # (paramsDict["inverse_beam"] and ' -inverse' or '')
                    lineToExecute = (
                        executable + endOfLineToExecute + " 2>&1 > /dev/null &"
                    )
                    logging.info(
                        "Process event %s, executing %s"
                        % (processEvent, str(lineToExecute))
                    )

                    subprocess.Popen(
                        str(lineToExecute),
                        shell=True,
                        stdin=None,
                        stdout=None,
                        stderr=None,
                        close_fds=True,
                    )
                else:
                    logging.getLogger().error(
                        "No program to execute found (%s)", executable
                    )
        except KeyError:
            logging.exception("autoprocessing: an error occurred")


def startInducedRadDam(datacollect_params, old={"xds_dir": None}):
    if not datacollect_params["xds_dir"] == old["xds_dir"]:
        old["xds_dir"] = datacollect_params["xds_dir"]
        eda_dirs = sorted(
            filter(
                os.path.isdir,
                [
                    os.path.join(datacollect_params["EDNA_files_dir"], x)
                    for x in os.listdir(datacollect_params["EDNA_files_dir"])
                    if x.startswith("EDA")
                ],
            )
        )
        EDApplication = eda_dirs[-1]
        os.system(
            "export TCL_LIBRARY=/usr/share/tcl8.4;/opt/pxsoft/bin/InducedRadDam.py -i -e %s -p %s &"
            % (EDApplication, datacollect_params["xds_dir"])
        )
    return True



from datetime import (
    datetime,
    timedelta,
)
from typing import (
    Dict,
    List,
    Optional,
)

from pydantic.v1 import (
    BaseModel,
    Field,
)


class Proposal(BaseModel):
    proposal_id: str = ""
    person_id: str = ""
    type: str = ""
    code: str = ""
    number: str = ""
    title: str = ""


class Lims(BaseModel):
    name: str = ""
    description: str = ""


class Session(BaseModel):
    session_id: str = ""
    beamline_name: str = ""
    start_date: str = ""  # YYYYMDD
    start_time: str = ""
    end_date: str = ""  # YYYYMDD
    end_time: str = ""

    # Proposal information
    title: str = ""
    code: str = ""
    number: str = ""
    proposal_id: str = ""
    proposal_name: str = ""

    comments: Optional[str] = ""

    start_datetime: datetime = Field(default_factory=datetime.now)
    end_datetime: Optional[datetime] = Field(
        default_factory=lambda: datetime.now() + timedelta(days=1)
    )

    # If rescheduled the actual dates are used instead
    actual_start_date: str = ""
    actual_start_time: str = ""
    actual_end_date: str = ""
    actual_end_time: str = ""

    nb_shifts: str = ""
    scheduled: str = ""

    # status of the session depending on wether it has been rescheduled or moved
    is_rescheduled: bool = False
    is_scheduled_time: bool = False
    is_scheduled_beamline: bool = False

    # direct links to different services
    user_portal_URL: Optional[str] = None
    data_portal_URL: Optional[str] = None
    logbook_URL: Optional[str] = None


class LimsUser(BaseModel):
    user_name: str = ""
    sessions: Optional[List[Session]] = []


class LimsSessionManager(BaseModel):
    active_session: Optional[Session] = None
    sessions: Optional[List[Session]] = []
    users: Optional[Dict[str, LimsUser]] = {}



"""
Class for cameras connected by EPICS Area Detector
"""
import logging
import os
from io import BytesIO
from threading import Thread

import gevent
import numpy as np
from PIL import Image

from mxcubecore import BaseHardwareObjects

CAMERA_DATA = "epicsCameraSample_data"
CAMERA_BACK = "epicsCameraSample_back"
CAMERA_EN_BACK = "epicsCameraSample_en_back"
CAMERA_ACQ_START = "epicsCameraSample_acq_start"
CAMERA_ACQ_STOP = "epicsCameraSample_acq_stop"
CAMERA_GAIN = "epicsCameraSample_gain"
CAMERA_GAIN_RBV = "epicsCameraSample_gain_rbv"
CAMERA_AUTO_GAIN = "epicsCameraSample_auto_gain"
CAMERA_AUTO_GAIN_RBV = "epicsCameraSample_auto_gain_rbv"
CAMERA_FPS_RBV = "epicsCameraSample_frames_per_second_rbv"
CAMERA_ACQ_TIME = "epicsCameraSample_acq_time"
CAMERA_ACQ_TIME_RBV = "epicsCameraSample_acq_time_rbv"
CAMERA_IMG_PIXEL_SIZE = "epicsCameraSample_img_pixel_size"
CAMERA_IMG_WIDTH = "epicsCameraSample_img_width"
CAMERA_IMG_HEIGHT = "epicsCameraSample_img_height"

class LNLSCamera(BaseHardwareObjects.HardwareObject):

    def __init__(self,name):
        BaseHardwareObjects.Device.__init__(self,name)
        self.liveState = False
        self.refreshing = False
        self.imagegen = None
        self.refreshgen = None
        self.imgArray = None
        self.qImage = None
        self.qImageHalf = None
        self.delay = None
        self.array_size = None
        # Status (cam is getting images)
        # This flag makes errors to be printed only when needed in the log,
        # which prevents the log file to get gigantic.
        self._print_cam_sucess = True
        self._print_cam_error_null = True
        self._print_cam_error_size = True
        self._print_cam_error_format = True

    def _init(self):
        self.stream_hash = "#"
        self.setIsReady(True)

    def init(self):
        self.read_sizes()
        # Start camera image acquisition
        self.setLive(True)
        # Snapshot
        self.centring_status = {"valid": False}
        self.snapshots_procedure = None

    def read_sizes(self):
        self.pixel_size = self.read_pixel_size()
        self.width = self.read_width()
        self.height = self.read_height()
        self.array_size = self.read_array_size()

    def poll(self):
        logging.getLogger("HWR").debug('LNLS Camera image acquiring has started.')
        self.imageGenerator(self.delay)

    def imageGenerator(self, delay):
        while self.liveState:
            self.getCameraImage()
            gevent.sleep(delay)
        logging.getLogger("HWR").debug('LNLS Camera image acquiring has stopped.')

    def getCameraImage(self):
        # Get the image from uEye camera IOC
        self.imgArray = self.get_channel_value(CAMERA_DATA)
        if self.imgArray is None:
            if self._print_cam_error_null:
                logging.getLogger("HWR").error("%s - Error: null camera image!" % (self.__class__.__name__))
                self._print_cam_sucess = True
                self._print_cam_error_null = False
                self._print_cam_error_size = True
                self._print_cam_error_format = True
            # Return error for this frame, but cam remains live for new frames
            return -1

        if len(self.imgArray) != self.array_size:
            # PS: This check possibly can be removed and the treatment can be
            # moved into the except scope.
            if self._print_cam_error_size:
                logging.getLogger("HWR").error(\
                "%s - Error in array lenght! Expected %d, but got %d." % \
                (self.__class__.__name__, self.array_size, len(self.imgArray)))
                self._print_cam_sucess = True
                self._print_cam_error_null = True
                self._print_cam_error_size = False
                self._print_cam_error_format = True
            # Try to read sizes again
            self.read_sizes()
            # Return error for this frame, but cam remains live for new frames
            return -1

        if self.refreshing:
            logging.getLogger("user_level_log").info("Camera was refreshed!")
            self.refreshing = False

        try:
            # Get data
            data = self.imgArray
            arr = np.array(data).reshape(self.height, self.width, self.pixel_size)
            # Convert data to rgb image
            img = Image.fromarray(arr)
            #img_rot = img.rotate(angle=0, expand=True)
            img_rgb = img.convert('RGB')
            # Get binary image
            with BytesIO() as f:
                img_rgb.save(f, format='JPEG')
                f.seek(0)
                img_bin_str = f.getvalue()
            # Sent image to gui
            self.emit("imageReceived", img_bin_str, self.height, self.width)
            #logging.getLogger("HWR").debug('Got camera image: ' + \
            #str(img_bin_str[0:10]))
            if self._print_cam_sucess:
                logging.getLogger("HWR").info("LNLSCamera is emitting images! Cam routine is ok.")
                self._print_cam_sucess = False
                self._print_cam_error_null = True
                self._print_cam_error_size = True
                self._print_cam_error_format = True
            return 0
        except:
            if self._print_cam_error_format:
                logging.getLogger("HWR").error('Error while formatting camera image')
                self._print_cam_sucess = True
                self._print_cam_error_null = True
                self._print_cam_error_size = True
                self._print_cam_error_format = False
            return -1

    def read_pixel_size(self):
        pixel_size = 1
        try:
            pixel_size = self.get_channel_value(CAMERA_IMG_PIXEL_SIZE)
            if pixel_size is None or pixel_size <= 0:
                pixel_size = 1
        except:
            print("Error on getting camera pixel size.")
        finally:
            logging.getLogger("HWR").info("LNLSCamera pixel size is %d." % (pixel_size))
            return pixel_size

    def read_width(self):
        width = 0
        try:
            width = self.get_channel_value(CAMERA_IMG_WIDTH)
            if width is None:
                width = 0
        except:
            print("Error on getting camera width.")
        finally:
            logging.getLogger("HWR").info("LNLSCamera width is %d." % (width))
            return width

    def read_height(self):
        height = 0
        try:
            height = self.get_channel_value(CAMERA_IMG_HEIGHT)
            if height is None:
                height = 0
        except:
            print("Error on getting camera height.")
        finally:
            logging.getLogger("HWR").info("LNLSCamera height is %d." % (height))
            return height

    def read_array_size(self):
        array_size = -1
        try:
            pixel_size = self.read_pixel_size()
            width = self.read_width()
            height = self.read_height()
            array_size = pixel_size * width * height
        except:
            print("Error on getting camera array size.")
        return array_size

    def get_pixel_size(self):
        return self.pixel_size

    def get_width(self):
        return self.width

    def get_height(self):
        return self.height

    def get_array_size(self):
        return self.array_size

    def getStaticImage(self):
        pass
        #qtPixMap = QtGui.QPixmap(self.source, "1")
        #self.emit("imageReceived", qtPixMap)

    def get_image_dimensions(self):
        return self.get_array_size()

    def getWidth(self):
        # X
        return self.get_width()

    def getHeight(self):
        # Z
        return self.get_height()

    def contrastExists(self):
        return False

    def brightnessExists(self):
        return False

    def gainExists(self):
        return True

    def get_gain(self):
        gain = None

        try:
            gain = self.get_channel_value(CAMERA_GAIN_RBV)
        except:
            print("Error getting gain of camera...")

        return gain

    def set_gain(self, gain):
        try:
            self.setValue(CAMERA_GAIN, gain)
        except:
            print("Error setting gain of camera...")

    def get_gain_auto(self):
        auto = None

        try:
            auto = self.get_channel_value(CAMERA_AUTO_GAIN_RBV)
        except:
            print("Error getting auto-gain of camera...")

        return auto

    def set_gain_auto(self, auto):
        try:
            self.setValue(CAMERA_AUTO_GAIN, auto)
        except:
            print("Error setting auto-gain of camera...")


    def get_exposure_time(self):
        exp = None

        try:
            exp = self.get_channel_value(CAMERA_ACQ_TIME_RBV)
        except:
            print("Error getting exposure time of camera...")

        return exp

    def set_exposure_time(self, exp):
        try:
            self.setValue(CAMERA_ACQ_TIME, exp)
        except:
            print("Error setting exposure time of camera...")

    def start_camera(self):
        try:
            self.setValue(CAMERA_BACK, 1)
            self.setValue(CAMERA_EN_BACK, 1)
            self.setValue(CAMERA_ACQ_STOP, 0)
            self.setValue(CAMERA_ACQ_START, 1)
        except:
            pass

    def stop_camera(self):
        try:
            self.setValue(CAMERA_ACQ_START, 0)
            self.setValue(CAMERA_ACQ_STOP, 1)
        except:
            pass

    def refresh_camera_procedure(self):
        self.refreshing = True

        # Try to reconnect to PVs
        self.reconnect(CAMERA_DATA)
        self.reconnect(CAMERA_IMG_PIXEL_SIZE)
        self.reconnect(CAMERA_IMG_WIDTH)
        self.reconnect(CAMERA_IMG_HEIGHT)
        self.reconnect(CAMERA_BACK)
        self.reconnect(CAMERA_EN_BACK)
        self.reconnect(CAMERA_ACQ_START)
        self.reconnect(CAMERA_ACQ_STOP)
        self.reconnect(CAMERA_GAIN)
        self.reconnect(CAMERA_GAIN_RBV)
        self.reconnect(CAMERA_AUTO_GAIN)
        self.reconnect(CAMERA_AUTO_GAIN_RBV)
        self.reconnect(CAMERA_FPS_RBV)
        self.reconnect(CAMERA_ACQ_TIME)
        self.reconnect(CAMERA_ACQ_TIME_RBV)

        # Try to stop camera image acquisition
        self.setLive(False)
        # Wait a while
        gevent.sleep(0.2)
        # Set PVs to start
        self.start_camera()
        # (Re)start camera image acquisition
        self.setLive(True)

    def refresh_camera(self):
        logging.getLogger("user_level_log").error("Resetting camera, please, wait a while...")
        print("refresh_camera")

        # Start a new thread to don't freeze UI
        self.refreshgen = gevent.spawn(self.refresh_camera_procedure)

    def setLive(self, live):
        try:
            if live and self.liveState == live:
                return

            self.liveState = live

            if live:
                logging.getLogger("HWR").info("LNLSCamera is going to poll images")
                self.delay = float(int(self.getProperty("interval"))/1000.0)
                thread = Thread(target=self.poll)
                thread.daemon = True
                thread.start()
            else:
                self.stop_camera()

            return True
        except:
            return False

    def takeSnapshot(self, *args):
        pass
        #imgFile = QtCore.QFile(args[0])
        #imgFile.open(QtCore.QIODevice.WriteOnly)
        #self.qtPixMap.save(imgFile,"PNG")
        #imgFile.close()

    def take_snapshots_procedure(self, image_count, snapshotFilePath, snapshotFilePrefix, logFilePath, runNumber, collectStart, collectEnd, motorHwobj, detectorHwobj):
        """
        Descript. : It takes snapshots of sample camera and camserver execution.
        """
        # Avoiding a processing of AbstractMultiCollect class for saving snapshots
        #centred_images = []
        centred_images = None
        positions = []

        try:
            # Calculate goniometer positions where to take snapshots
            if (collectEnd is not None and collectStart is not None):
                interval = (collectEnd - collectStart)
            else:
                interval = 0

            # To increment in angle increment
            increment = 0 if ((image_count -1) == 0) else (interval / (image_count -1))

            for incrementPos in range(image_count):
                if (collectStart is not None):
                    positions.append(collectStart + (incrementPos * increment))
                else:
                    positions.append(motorHwobj.getPosition())

            # Create folders if not found
            if (not os.path.exists(snapshotFilePath)):
                try:
                    os.makedirs(snapshotFilePath, mode=0o700)
                except OSError as diag:
                    logging.getLogger().error("Snapshot: error trying to create the directory %s (%s)" % (snapshotFilePath, str(diag)))

            for index in range(image_count):
                while (motorHwobj.getPosition() < positions[index]):
                    gevent.sleep(0.02)

                logging.getLogger("HWR").info("%s - taking snapshot #%d" % (self.__class__.__name__, index + 1))

                # Save snapshot image file
                imageFileName = os.path.join(snapshotFilePath, snapshotFilePrefix + "_" + str(round(motorHwobj.getPosition(),2)) + "_" + motorHwobj.getEgu() + "_snapshot.png")

                #imageInfo = self.takeSnapshot(imageFileName)

                # This way all shapes will be also saved...
                self.emit("saveSnapshot", imageFileName)

                # Send a command to detector hardware-object to take snapshot of camserver execution...
                if (logFilePath and detectorHwobj):
                    detectorHwobj.takeScreenshotOfXpraRunningProcess(image_path=logFilePath, run_number=runNumber)

                #centred_images.append((0, str(imageInfo)))
                #centred_images.reverse()
        except:
            logging.getLogger("HWR").exception("%s - could not take crystal snapshots" % (self.__class__.__name__))

        return centred_images

    def take_snapshots(self, image_count, snapshotFilePath, snapshotFilePrefix, logFilePath, runNumber, collectStart, collectEnd, motorHwobj, detectorHwobj, wait=False):
        """
        Descript. :  It takes snapshots of sample camera and camserver execution.
        """
        if image_count > 0:
            self.snapshots_procedure = gevent.spawn(self.take_snapshots_procedure, image_count, snapshotFilePath, snapshotFilePrefix, logFilePath, runNumber, collectStart, collectEnd, motorHwobj, detectorHwobj)

            self.centring_status["images"] = []

            self.snapshots_procedure.link(self.snapshots_done)

            if wait:
                self.centring_status["images"] = self.snapshots_procedure.get()

    def snapshots_done(self, snapshots_procedure):
        """
        Descript. :
        """
        try:
            self.centring_status["images"] = snapshots_procedure.get()
        except:
            logging.getLogger("HWR").exception("%s - could not take crystal snapshots" % (self.__class__.__name__))

    def cancel_snapshot(self):
        try:
            self.snapshots_procedure.kill()
        except:
            pass

    def __del__(self):
        logging.getLogger().exception("%s - __del__()!" % (self.__class__.__name__))
        self.stop_camera()
        self.setLive(False)



#  Project: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""
Read the state of the hutch from the PSS device server and take actions
when enter (1) or interlock (0) the hutch.
0 = The hutch has been interlocked and the sample environment should be made
     ready for data collection. The actions are extract the detector cover,
     move the detector to its previous position, set the MD2 to Centring.
1 = The interlock is cleared and the user is entering the hutch to change
      the sample(s). The actions are insert the detector cover, move the
      detecto to a safe position, set MD2 to sample Transfer.
Example xml file:
<object class = "ESRF.BlissHutchTrigger">
  <username>Hutch Trigger</username>
  <pss_tango_device>acs:10000/bl/sa-pss/id30-crate02</pss_tango_device>
  <polling_interval>5</polling_interval>
  <pss_card_ch>9/4</pss_card_ch>
  <object href="/bliss" role="controller"/>
  <values>{"ENABLED": 1, "DISABLED": 0}</values>
</object>
"""

import logging

from gevent import (
    sleep,
    spawn,
)
from PyTango import DevFailed
from PyTango.gevent import DeviceProxy

from mxcubecore.HardwareObjects.abstract.AbstractNState import AbstractNState

__copyright__ = """ Copyright © 2010-2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class BlissHutchTrigger(AbstractNState):
    """Read the state of the hutch from the PSS and take actions."""

    def __init__(self, name):
        super(BlissHutchTrigger, self).__init__(name)
        self._bliss_obj = None
        self._proxy = None
        self.card = None
        self.channel = None
        self._pss_value = None
        self._nominal_value = None
        self._polling_interval = None
        self._poll_task = None

    def init(self):
        """Initialise properties and polling"""
        super(BlissHutchTrigger, self).init()
        self._bliss_obj = self.get_object_by_role("controller")
        tango_device = self.get_property("pss_tango_device")
        try:
            self._proxy = DeviceProxy(tango_device)
        except DevFailed as _traceback:
            last_error = _traceback[-1]
            msg = f"{self.name()}: {last_error['desc']}"
            raise RuntimeError(msg)

        pss = self.get_property("pss_card_ch")
        try:
            self.card, self.channel = map(int, pss.split("/"))
        except AttributeError:
            msg = f"{self.name()}: cannot find PSS number"
            raise RuntimeError(msg)

        # polling interval [s]
        self._polling_interval = self.get_property("polling_interval", 5)
        self._pss_value = self.get_pss_value()
        # enable by default
        self.update_value(self.VALUES.ENABLED)
        self._poll_task = spawn(self._do_polling)

    def _do_polling(self):
        """Do the polling of the PSS system"""
        while True:
            self._update_value(self.get_pss_value())
            sleep(self._polling_interval)

    def get_value(self):
        """The value corresponds to activate/deactivate the hutch trigger
        polling.
        Returns:
            (ValueEnum): Last set value.
        """
        return self._nominal_value

    def set_value(self, value, timeout=0):
        super(BlissHutchTrigger, self).set_value(value, timeout=0)

    def _set_value(self, value):
        """Set the hutch trigger enable/disable value
        Args:
            value (ValueEnum): ENABLED/DISABLED.
        """
        self._nominal_value = value
        self.emit("valueChanged", (value,))

    def get_pss_value(self):
        """Get the interlock value
        Returns:
            (bool): 0 = Hutch interlocked, 1 = Hutch not interlocked.
        """
        _ch1 = self._proxy.GetInterlockState([self.card - 1, 2 * (self.channel - 1)])[0]
        _ch2 = self._proxy.GetInterlockState(
            [self.card - 1, 2 * (self.channel - 1) + 1]
        )[0]
        return _ch1 & _ch2

    def _update_value(self, value=None):
        """Check if the pss value has changed (door opens/closes).
        Args:
            value (bool): The value
        """
        if self._nominal_value == self.VALUES.ENABLED:
            if value is None:
                value = self.get_pss_value()

            if self._pss_value != value:
                self._pss_value = value
                # now do the action
                self.hutch_actions(1 - value)

    def hutch_actions(self, enter, **kwargs):
        """Take action as function of the PSS state
        Args:
            enter(bool): True if entering hutch (interlock state = 1)
        """
        msg = "%s hutch" % ("Entering" if enter else "Leaving")
        logging.getLogger("user_level_log").info(msg)
        self._bliss_obj.hutch_actions(enter, hutch_trigger=True, **kwargs)



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""EMBLCollect - defines osc, helical and mesh collections"""

import os

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractCollect import AbstractCollect
from mxcubecore.TaskUtils import task

__credits__ = ["EMBL Hamburg"]
__category__ = "General"


class EMBLCollect(AbstractCollect):
    """Main data collection class. Inherited from AbstractCollect.
       Collection is done by setting collection parameters and
       executing collect command
    """

    def __init__(self, name):

        AbstractCollect.__init__(self, name)
        self._previous_collect_status = None
        self._actual_collect_status = None

        self._collect_frame = None
        self._exp_type_dict = {}
        self.break_bragg_released = False

        self.aborted_by_user = None
        self.run_autoprocessing = None

        self.chan_collect_status = None
        self.chan_collect_frame = None
        self.chan_collect_error = None
        self.chan_undulator_gap = None

        self.cmd_collect_compression = None
        self.cmd_collect_description = None
        self.cmd_collect_detector = None
        self.cmd_collect_directory = None
        self.cmd_collect_energy = None
        self.cmd_collect_exposure_time = None
        self.cmd_collect_helical_position = None
        self.cmd_collect_in_queue = None
        self.cmd_collect_images_per_trigger = None
        self.cmd_collect_num_images = None
        self.cmd_collect_overlap = None
        self.cmd_collect_processing = None
        self.cmd_collect_range = None
        self.cmd_collect_raster_lines = None
        self.cmd_collect_raster_range = None
        self.cmd_collect_resolution = None
        self.cmd_collect_scan_type = None
        self.cmd_collect_shutter = None
        self.cmd_collect_shutterless = None
        self.cmd_collect_start_angle = None
        self.cmd_collect_start_image = None
        self.cmd_collect_template = None
        self.cmd_collect_transmission = None
        self.cmd_collect_space_group = None
        self.cmd_collect_unit_cell = None
        self.cmd_collect_start = None
        self.cmd_collect_abort = None
        self.cmd_collect_xds_data_range = None

    def init(self):
        """Main init method"""

        AbstractCollect.init(self)

        self._exp_type_dict = {"Mesh": "raster", "Helical": "Helical"}

        self.chan_collect_status = self.get_channel_object("collectStatus")
        self._actual_collect_status = self.chan_collect_status.get_value()
        self.chan_collect_status.connect_signal("update", self.collect_status_update)
        self.chan_collect_frame = self.get_channel_object("collectFrame")
        self.chan_collect_frame.connect_signal("update", self.collect_frame_update)
        self.chan_collect_error = self.get_channel_object("collectError")
        self.chan_collect_error.connect_signal("update", self.collect_error_update)
        self.cmd_collect_compression = self.get_command_object("collectCompression")
        self.cmd_collect_description = self.get_command_object("collectDescription")
        self.cmd_collect_detector = self.get_command_object("collectDetector")
        self.cmd_collect_directory = self.get_command_object("collectDirectory")
        self.cmd_collect_energy = self.get_command_object("collectEnergy")
        self.cmd_collect_exposure_time = self.get_command_object("collectExposureTime")
        self.cmd_collect_images_per_trigger = self.get_command_object(
            "collectImagesPerTrigger"
        )
        self.cmd_collect_helical_position = self.get_command_object(
            "collectHelicalPosition"
        )
        self.cmd_collect_in_queue = self.get_command_object("collectInQueue")
        self.cmd_collect_num_images = self.get_command_object("collectNumImages")
        self.cmd_collect_overlap = self.get_command_object("collectOverlap")
        self.cmd_collect_processing = self.get_command_object("collectProcessing")
        self.cmd_collect_range = self.get_command_object("collectRange")
        self.cmd_collect_raster_lines = self.get_command_object("collectRasterLines")
        self.cmd_collect_raster_range = self.get_command_object("collectRasterRange")
        self.cmd_collect_resolution = self.get_command_object("collectResolution")
        self.cmd_collect_scan_type = self.get_command_object("collectScanType")
        self.cmd_collect_shutter = self.get_command_object("collectShutter")
        self.cmd_collect_shutterless = self.get_command_object("collectShutterless")
        self.cmd_collect_start_angle = self.get_command_object("collectStartAngle")
        self.cmd_collect_start_image = self.get_command_object("collectStartImage")
        self.cmd_collect_template = self.get_command_object("collectTemplate")
        self.cmd_collect_transmission = self.get_command_object("collectTransmission")
        self.cmd_collect_space_group = self.get_command_object("collectSpaceGroup")
        self.cmd_collect_unit_cell = self.get_command_object("collectUnitCell")
        self.cmd_collect_xds_data_range = self.get_command_object("collectXdsDataRange")

        self.cmd_collect_nexp_frame = self.get_command_object("collectImagesPerTrigger")

        self.cmd_collect_start = self.get_command_object("collectStart")
        self.cmd_collect_abort = self.get_command_object("collectAbort")

        self.emit("collectConnected", (True,))
        self.emit("collectReady", (True,))

    def data_collection_hook(self):
        """Main collection hook"""

        if self.aborted_by_user:
            self.collection_failed("Aborted by user")
            self.aborted_by_user = False
            return

        if self._actual_collect_status in [
            "ready",
            "unknown",
            "error",
            "not available",
        ]:
            HWR.beamline.diffractometer.save_centring_positions()
            comment = "Comment: %s" % str(
                self.current_dc_parameters.get("comments", "")
            )
            self._error_msg = ""
            self._collecting = True
            self._collect_frame = None

            osc_seq = self.current_dc_parameters["oscillation_sequence"][0]
            file_info = self.current_dc_parameters["fileinfo"]
            sample_ref = self.current_dc_parameters["sample_reference"]

            if HWR.beamline.image_tracking is not None:
                HWR.beamline.image_tracking.set_image_tracking_state(True)

            if self.cmd_collect_compression is not None:
                self.cmd_collect_compression(file_info["compression"])

            self.cmd_collect_description(comment)
            self.cmd_collect_detector(HWR.beamline.detector.get_collect_name())
            self.cmd_collect_directory(str(file_info["directory"]))
            self.cmd_collect_exposure_time(osc_seq["exposure_time"])
            self.cmd_collect_in_queue(self.current_dc_parameters["in_queue"] != False)
            self.cmd_collect_nexp_frame(1)
            self.cmd_collect_overlap(osc_seq["overlap"])

            shutter_name = HWR.beamline.detector.get_shutter_name()
            if shutter_name is not None:
                self.cmd_collect_shutter(shutter_name)

            if osc_seq["overlap"] == 0:
                self.cmd_collect_shutterless(1)
            else:
                self.cmd_collect_shutterless(0)

            self.cmd_collect_range(osc_seq["range"])
            if self.current_dc_parameters["experiment_type"] != "Mesh":
                self.cmd_collect_num_images(osc_seq["number_of_images"])

            if self.cmd_collect_processing is not None:
                self.cmd_collect_processing(True)
                # GB 2019030: no idea why this could be unset.....
                #    self.current_dc_parameters["processing_parallel"]
                #    in (True, "MeshScan", "XrayCentering")
                #

            #if self.current_dc_parameters["processing_online"] is False:
            #    self.cmd_collect_processing(False)

            # GB 2018-05-16 : Workaround a fuzzy mesh scan interface of MD3
            # if self.current_dc_parameters['experiment_type'] == 'Mesh':
            #   _do_start = osc_seq['start'] - 0.5 * osc_seq['range']*
            # osc_seq['number_of_images']/float(osc_seq['number_of_lines'])
            #   print _do_start, ' Here'
            # else:
            #   _do_start = osc_seq['start']

            self.cmd_collect_start_angle(osc_seq["start"])
            # self.cmd_collect_start_angle(_do_start)

            self.cmd_collect_start_image(osc_seq["start_image_number"])
            self.cmd_collect_template(str(file_info["template"]))
            space_group = str(sample_ref["spacegroup"])
            if len(space_group) == 0:
                space_group = " "
            self.cmd_collect_space_group(space_group)
            unit_cell = list(eval(sample_ref["cell"]))
            self.cmd_collect_unit_cell(unit_cell)

            if self.current_dc_parameters["experiment_type"] == "OSC":
                xds_range = (
                    osc_seq["start_image_number"],
                    osc_seq["start_image_number"] + osc_seq["number_of_images"] - 1,
                )
                self.cmd_collect_xds_data_range(xds_range)
            elif (
                self.current_dc_parameters["experiment_type"] == "Collect - Multiwedge"
            ):
                xds_range = self.current_dc_parameters["in_interleave"]
                self.cmd_collect_xds_data_range(xds_range)

            if osc_seq["num_triggers"] and osc_seq["num_images_per_trigger"]:
                self.cmd_collect_scan_type("still")
                self.cmd_collect_images_per_trigger(osc_seq["num_images_per_trigger"])
            else:
                self.cmd_collect_scan_type(
                    self._exp_type_dict.get(
                        self.current_dc_parameters["experiment_type"], "OSC"
                    )
                )
            self.cmd_collect_start()
        else:
            self.collection_failed(
                "Unable to start collection. "
                + "Detector server is in %s state" % self._actual_collect_status
            )

    def collect_status_update(self, status):
        """Status event that controls execution

        :param status: collection status
        :type status: string
        """
        if status != self._actual_collect_status:
            self._previous_collect_status = self._actual_collect_status
            self._actual_collect_status = status
            if self._collecting:
                if self._actual_collect_status == "error":
                    self.collection_failed()
                elif self._actual_collect_status == "collecting":
                    self._store_image_in_lims_by_frame_num(1)
                if self._previous_collect_status is None:
                    if self._actual_collect_status == "busy":
                        self.print_log("HWR", "info", "Collection: Preparing ...")
                elif self._previous_collect_status == "busy":
                    if self._actual_collect_status == "collecting":
                        self.emit("collectStarted", (None, 1))
                elif self._previous_collect_status == "collecting":
                    if self._actual_collect_status == "ready":
                        self.collection_finished()
                    elif self._actual_collect_status == "aborting":
                        self.print_log("HWR", "info", "Collection: Aborting...")
                        self.collection_failed()

    def collect_error_update(self, error_msg):
        """Collect error behaviour

        :param error_msg: error message
        :type error_msg: string
        """

        if self._collecting and len(error_msg) > 0:
            self._error_msg = error_msg
            self.print_log(
                "GUI", "error", "Collection: Error from detector server: %s" % error_msg
            )

    def collection_finished(self):
        """Additionaly sets break bragg if it was previously released"""
        AbstractCollect.collection_finished(self)
        if (
            self.current_dc_parameters["in_queue"] is False
            and self.break_bragg_released
        ):
            self.break_bragg_released = False
            HWR.beamline.energy.set_break_bragg()

    def collect_frame_update(self, frame):
        """Image frame update

        :param frame: frame num
        :type frame: int
        """
        if self._collecting:
            if self.current_dc_parameters["in_interleave"]:
                number_of_images = self.current_dc_parameters["in_interleave"][1]
            else:
                number_of_images = self.current_dc_parameters["oscillation_sequence"][
                    0
                ]["number_of_images"]
            if self._collect_frame != frame:
                self._collect_frame = frame
                self.emit("progressStep", (int(float(frame) / number_of_images * 100)))
                self.emit("collectImageTaken", frame)

    def _store_image_in_lims_by_frame_num(self, frame_number):
        """Store image in lims

        :param frame: dict with frame parameters
        :type frame: dict
        :param motor_position_id: position id
        :type motor_position_id: int
        """
        self.trigger_auto_processing("image", frame_number)
        return self._store_image_in_lims(frame_number)

    def trigger_auto_processing(self, process_event, frame_number):
        """Starts autoprocessing"""
        HWR.beamline.offline_processing.execute_autoprocessing(
            process_event,
            self.current_dc_parameters,
            frame_number,
            self.current_dc_parameters["processing_offline"],
        )

    def stop_collect(self):
        """Stops collect"""

        AbstractCollect.stop_collect(self)

        self.cmd_collect_abort()
        HWR.beamline.detector.close_cover()

    def set_helical_pos(self, arg):
        """Sets helical positions
           8 floats describe:
             p1AlignmY, p1AlignmZ, p1CentrX, p1CentrY
             p2AlignmY, p2AlignmZ, p2CentrX, p2CentrY
        """
        helical_positions = [
            arg["1"]["phiy"],
            arg["1"]["phiz"],
            arg["1"]["sampx"],
            arg["1"]["sampy"],
            arg["2"]["phiy"],
            arg["2"]["phiz"],
            arg["2"]["sampx"],
            arg["2"]["sampy"],
        ]
        self.cmd_collect_helical_position(helical_positions)

    def set_mesh_scan_parameters(
        self, num_lines, num_total_frames, mesh_center, mesh_range
    ):
        """Sets mesh parameters"""
        self.cmd_collect_raster_lines(num_lines)
        self.cmd_collect_num_images(num_total_frames / num_lines)
        # GB collection server interface assumes the order: fast, slow for mesh
        # range. Need reversal at P14:
        if HWR.beamline.session.beamline_name == "P13":
            self.cmd_collect_raster_range(mesh_range)
        else:
            self.cmd_collect_raster_range(mesh_range[::-1])

    @task
    def _take_crystal_snapshot(self, snapshot_filename):
        """Saves crystal snapshot"""
        HWR.beamline.sample_view.save_scene_snapshot(snapshot_filename)

    @task
    def _take_crystal_animation(self, animation_filename, duration_sec=1):
        """Rotates sample by 360 and composes a gif file
           Animation is saved as the fourth snapshot
        """

        HWR.beamline.sample_view.save_scene_animation(animation_filename, duration_sec)

    # def set_energy(self, value):
    #     """Sets energy"""
    #     """
    #     if abs(value - self.get_energy()) > 0.005 and not self.break_bragg_released:
    #         self.break_bragg_released = True
    #         if hasattr(HWR.beamline.energy, "release_break_bragg"):
    #             HWR.beamline.energy.release_break_bragg()
    #         self.cmd_collect_energy(value * 1000.0)
    #     else:
    #     """
    #     self.cmd_collect_energy(self.get_energy() * 1000)

    def set_resolution(self, value):
         """Sets resolution in A"""
         if not value:
             value = self.get_resolution()
         self.cmd_collect_resolution(value)

    def set_transmission(self, value):
         """Sets transmission in %"""
         self.cmd_collect_transmission(value)

    @task
    def move_motors(self, motor_position_dict):
        """Move to centred position"""
        HWR.beamline.diffractometer.move_motors(motor_position_dict)

    def prepare_input_files(self):
        """Prepares xds directory"""
        i = 1
        while True:
            xds_input_file_dirname = "xds_%s_%s_%d" % (
                self.current_dc_parameters["fileinfo"]["prefix"],
                self.current_dc_parameters["fileinfo"]["run_number"],
                i,
            )
            xds_directory = os.path.join(
                self.current_dc_parameters["fileinfo"]["process_directory"],
                xds_input_file_dirname,
            )
            if not os.path.exists(xds_directory):
                break
            i += 1

        self.current_dc_parameters["xds_dir"] = xds_directory

        return xds_directory, ""

    def get_undulators_gaps(self):
        """Return triplet with gaps. In our case we have one gap,
        """
        und_gaps = {}
        if self.chan_undulator_gap:
            und_gaps = self.chan_undulator_gap.get_value()
            if not isinstance(und_gaps, (list, tuple)):
                und_gaps = list(und_gaps)
        return und_gaps

    def get_machine_current(self):
        """Returns flux"""
        return HWR.beamline.machine_info.get_current()

    def get_machine_message(self):
        """Returns machine message"""
        return HWR.beamline.machine_info.get_message()

    def get_machine_fill_mode(self):
        """Returns machine filling mode"""
        fill_mode = str(HWR.beamline.machine_info.get_message())
        return fill_mode[:20]

    def get_beamline_configuration(self, *args):
        """Returns beamline config"""
        return self.bl_config._asdict()

    def get_total_absorbed_dose(self):
        return float("%.3e" % HWR.beamline.flux.get_total_absorbed_dose())

    def set_run_autoprocessing(self, status):
        """Enables or disables autoprocessing after a collection"""
        self.run_autoprocessing = status

    def create_file_directories(self):
        self.prepare_input_files()
